# -*- coding: utf-8 -*-
"""MOFFragmentor aims to provide Python abstractions for reticular chemistry."""
__version__ = "0.0.1"
from loguru import logger

from .mof import MOF  # noqa: F401

logger.disable("moffragmentor")



# -*- coding: utf-8 -*-
"""Test the branching point location"""
from moffragmentor.fragmentor.branching_points import (
    _is_branch_point,
    filter_branch_points,
    get_branch_points,
)


def test_get_branch_points(abaxin):
    """Test the branch point detection on a case that previously failed (#70)."""
    bp = get_branch_points(abaxin)
    assert set(bp) == {146, 147, 72, 73, 112, 113, 52, 53, 92, 93}


def test_is_branch_point(get_agn_mof):
    """In this case, all except for the Ag is a branch point candidate."""
    bp = [i for i, _ in enumerate(get_agn_mof.structure) if _is_branch_point(get_agn_mof, i)]
    assert len(bp) == 144

    filtered_bp = filter_branch_points(get_agn_mof, bp)
    assert len(filtered_bp) == 46
    for i in filtered_bp:
        assert get_agn_mof.structure[i].specie.symbol == "N"



# -*- coding: utf-8 -*-
"""Defines Python representations for nets and voltage.

The VoltageEdge class implementation is based on https://github.com/qai222/CrystalGraph,
which is licensed under the MIT license provided here:

MIT License

Copyright (c) 2021 Alex Ai

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from collections import OrderedDict, defaultdict
from typing import Dict, Iterable, List, Optional, Tuple, Union

import networkx as nx
import numpy as np
from backports.cached_property import cached_property
from loguru import logger
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.core import IMolecule, Lattice, Molecule, Structure
from pymatgen.util.coord import pbc_diff

from moffragmentor.sbu import SBU
from moffragmentor.sbu.linkercollection import LinkerCollection
from moffragmentor.sbu.nodecollection import NodeCollection

from .utils.systre import _get_systre_input_from_pmg_structure_graph, run_systre


def is_3d_parallel(v1: np.array, v2: np.array, eps: float = 1e-5) -> bool:
    if np.allclose(v1, 0) and not np.allclose(v2, 0):
        return False
    if np.allclose(v2, 0) and not np.allclose(v1, 0):
        return False
    cp = np.cross(v1, v2)
    return np.allclose(cp, np.zeros(3), atol=eps)


def get_image_if_on_edge(frac_coords):
    # exactly one --> return the -1 image
    images = np.zeros(3, dtype=int)
    is_exactly_one = np.argwhere(np.abs(frac_coords - 1) < 1e-4)
    if is_exactly_one.size > 0:
        images[is_exactly_one] = -1
    is_exactly_zero = np.argwhere(np.abs(frac_coords) < 1e-4)
    if is_exactly_zero.size > 0:
        images[is_exactly_zero] = 1
    return images


def sanitize_graph(structure_graph):
    # if there is a node with only one edge,
    # it is probably because the binding partner is at an edge or corner of the structure
    # that is, at least one fractional coordinate is 0 or 1
    # in this case, we add a bond to the image of this site
    new_edges = []
    for i, _ in enumerate(structure_graph.structure):
        connected_sites = structure_graph.get_connected_sites(i)
        if len(connected_sites) == 1:
            frac_coords = connected_sites[0].site.frac_coords
            images = get_image_if_on_edge(frac_coords)
            if np.any(images != 0):
                new_edges.append((i, connected_sites[0].index, images))

    for edge in new_edges:
        structure_graph.add_edge(edge[0], edge[1], to_jimage=edge[2])


class VoltageEdge:
    """Representation of an edge in a labeled quotient graph (LQG).

    This class simplifies the construction of LQGs (via the equiality operator).
    """

    def __init__(self, vector: np.ndarray, n1: int, n2: int, n1_image: tuple, n2_image: tuple):
        """Initialize a VoltageEdge.

        Args:
            vector (np.ndarray): The vector of the edge
                (difference between coordinates of nodes).
            n1 (int): The index of the first node.
            n2 (int): The index of the second node.
            n1_image (tuple): The image of the first node.
            n2_image (tuple): The image of the second node.
        """
        self.vector = vector
        self.n1 = n1
        self.n2 = n2
        self.n1_image = n1_image if n1_image is not None else (0, 0, 0)
        self.n2_image = n2_image if n2_image is not None else (0, 0, 0)
        self.terminals = frozenset([self.n1, self.n2])

    def __repr__(self):
        """Return a string representation of the VoltageEdge."""
        terms = (self.n1, self.n2)
        imags = (self.n1_image, self.n2_image)
        a, b = ((x, t) for x, t in sorted(zip(imags, terms), key=lambda x: x[1]))
        return "Edge: {}, voltage: {}, vector: {}".format((a, b), self.voltage, self.vector)

    @property
    def length(self) -> float:
        """Length of the edge."""
        return np.linalg.norm(self.vector)

    @property
    def voltage(self) -> Tuple[int, int, int]:
        """Voltage is the tuple describing the direction of the edge.

        In simple words, it represents the translation operation.

        Returns:
            Tuple[int, int, int]: The voltage of the edge.
        """
        terms = (self.n1, self.n2)
        imags = (self.n1_image, self.n2_image)
        a_image, b_image = (x for x, _ in sorted(zip(imags, terms), key=lambda x: x[1]))
        return tuple(a_image[i] - b_image[i] for i in range(3))

    def __eq__(self, other):
        """Check if two VoltageEdges are equal."""
        eps = 1e-5
        is_parallel = is_3d_parallel(self.vector, other.vector, eps=eps)
        if not is_parallel:
            return False
        is_sameterminal = self.terminals == other.terminals
        if not is_sameterminal:
            return False
        is_eqlength = abs(self.length - other.length) < eps
        if not is_eqlength:
            return False
        is_eqvoltage = is_3d_parallel(
            self.voltage, other.voltage
        ) or self.voltage == other.voltage == (0, 0, 0)
        if not is_eqvoltage:
            return False
        return True


class NetNode:
    def __init__(
        self,
        coords: np.array,
        branching_coord_in_cell: Optional[Iterable[np.array]] = None,
        all_branching_coord: Optional[Iterable[np.array]] = None,
        molecule: Optional[Union[Molecule, IMolecule]] = None,
        flavor: Optional[str] = None,
    ):
        """Initialize a NetNode.

        Args:
            coords (np.array): Coordinates of the node (center of mass).
            branching_coord_in_cell (Optional[Iterable[np.array]], optional):
                Coordinates of the branching indices that are in the original unit cell.
                Defaults to None.
            all_branching_coord (Optional[Iterable[np.array]], optional):
                Coordinates of all branching sites. Defaults to None.
            molecule (Optional[Union[Molecule, IMolecule]], optional):
                Pymatgen molecule object. Defaults to None.
            flavor (Optional[str], optional):
                String describing the net type, e.g. "linker", "metalcluster. Defaults to None.
        """
        self._coords = coords
        self._molecule = molecule
        self._flavor = flavor
        self._branching_coord_in_cell = branching_coord_in_cell
        self._all_branching_coord = all_branching_coord

    def __eq__(self, other: "NetNode") -> bool:  # noqa: F821 - forward reference
        """Check if two NetNodes are equal.

        Two NetNodes are equal if they have the same coordinates.

        Args:
            other (NetNode): The other NetNode to compare to.

        Returns:
            bool: True if the two NetNodes are equal, False otherwise.
        """
        return self._coords == other._coords

    def is_periodic_image(
        self, other: "NetNode", lattice: Lattice, tolerance: float = 1e-8
    ):  # noqa: F821 - forward reference
        """
        Return `True` if the two nodes are periodic images of each other.

        Args:
            other ("NetNode"): NetNode
            lattice (Lattice): The lattice of the crystal.
            tolerance (float): The tolerance for the distance between the two nodes.

        Returns:
            A boolean value.
        """
        return is_periodic_image(self._coords, other._coords, lattice, tolerance)

    def periodic_image(self, other: "NetNode", lattice: Lattice):  # noqa: F821 - forward reference
        """`periodic_image` returns the periodic image of `other` w.r.t to `self`.

        Args:
            other ("NetNode"): NetNode
            lattice (Lattice): The lattice of the crystal.

        Returns:
            np.array: The periodic image of the other node in the lattice.

        Raises:
            ValueError: If the two nodes are not periodic images of each other.
        """
        if not self.is_periodic_image(other, lattice):
            raise ValueError("Nodes are not periodic images")
        frac_self = lattice.get_fractional_coords(self._coords)
        frac_other = lattice.get_fractional_coords(other._coords)
        return which_periodic_image(frac_self, frac_other, lattice)[-1][0][1]

    def __repr__(self) -> str:
        """Return a string representation for a NetNode."""
        return "NetNode: {:.2f} {:.2f} {:.2f} {}".format(
            self._coords[0], self._coords[1], self._coords[2], self._flavor
        )


def is_periodic_image(
    a: np.array, b: np.array, lattice: Lattice, tolerance: float = 1e-8, is_frac: bool = False
) -> bool:
    """
    Return `True` if `a` and `b` are periodic images of each other.

    Args:
        a (np.array): np.array, b: np.array: The two points you want to compare.
        b (np.array): np.array
        lattice (Lattice): The lattice of the structure.
        tolerance (float): The tolerance for the difference between
            the fractional coordinates of the two points.
        is_frac (bool): Whether the points are in fractional coordinates.

    Returns:
        bool: The function is_periodic_image returns a boolean value.
    """
    if not is_frac:
        a_frac_coords = lattice.get_fractional_coords(a)
        b_frac_coords = lattice.get_fractional_coords(b)
    else:
        a_frac_coords = a
        b_frac_coords = b
    frac_diff = pbc_diff(a_frac_coords, b_frac_coords)
    return np.allclose(frac_diff, [0, 0, 0], atol=tolerance)


class Net:
    def __init__(
        self,
        nodes: Dict[str, Tuple[NetNode, int]],
        edges: Iterable[VoltageEdge],
        lattice: Lattice,
    ):
        """Initialize a Net object.

        Args:
            nodes (Dict[str, Tuple[NetNode, int]]): A dictionary of NetNode objects.
            edges (Iterable[VoltageEdge]): An iterable of VoltageEdge objects.
            lattice (Lattice): The lattice of the crystal.
        """
        self.nodes = nodes
        self.edges = edges
        self.lattice = lattice

    def get_dummy_structure(self) -> Structure:
        coords = []
        symbols = []
        for node, _ in self.nodes.values():
            if "metal" in node._flavor:
                coords.append(node._coords)
                symbols.append("Si")
            else:
                coords.append(node._coords)
                symbols.append("O")

        return Structure(
            lattice=self.lattice, species=symbols, coords=coords, coords_are_cartesian=True
        )

    def get_pmg_structure_graph(
        self, simplify: bool = True, sanitize: bool = False
    ) -> StructureGraph:
        """Return a StructureGraph object from the Net object

        Args:
            simplify (bool): Whether to simplify the graph. Defaults to True.
                If True, 2c vertices are removed.
            sanitize (bool): Whether to sanitize the graph. Defaults to False.
                If True, add edge to periodic image for sites with only one neighbor
                if this neighbor is at the egde of the unit cell.

        Returns:
            StructureGraph: The StructureGraph object.
        """
        s = self.get_dummy_structure()
        edge_dict = {}

        for edge in self.edges:
            edge_dict[
                (
                    edge.n1,
                    edge.n2,
                    (edge.n1_image[0], edge.n1_image[1], edge.n1_image[2]),
                    (edge.n2_image[0], edge.n2_image[1], edge.n2_image[2]),
                )
            ] = None

        structure_graph = StructureGraph.with_edges(s, edge_dict)
        if sanitize:
            sanitize_graph(structure_graph)
        if simplify:
            structure_graph = _simplify_structure_graph(structure_graph)
        return structure_graph

    @cached_property
    def _systre_results(self) -> dict:
        """Return the RCSR code of the net."""
        return run_systre(
            _get_systre_input_from_pmg_structure_graph(self.get_pmg_structure_graph(), self.lattice)
        )

    @property
    def rcsr_code(self) -> str:
        """Return the RCSR code of the net."""
        try:
            return self._systre_results["rcsr_code"]
        except TypeError:
            return ""

    @property
    def p(self) -> int:
        return len(self._systre_results["relaxed_node_positions"])

    @property
    def q(self) -> int:
        return len(self._systre_results["relaxed_edge_positions"])

    def _construct_nx_multigraph(self) -> nx.MultiGraph:
        graph = nx.MultiGraph()
        for edge in self.edges:
            graph.add_edge(edge.n1, edge.n2, voltage=edge.voltage, vector=edge.vector)
        return graph

    @property
    def _cns(self) -> Dict[str, int]:
        """Return a list of the number of nodes in each edge."""
        return {k: len(v) for k, v in self._edge_dict.items()}

    @property
    def _edge_dict(self) -> Dict[str, List[VoltageEdge]]:
        edge_dict = defaultdict(list)
        for edge in self.edges:
            edge_dict[edge.n1].append(edge)
            edge_dict[edge.n2].append(edge)
        return edge_dict


def add_node_to_collection(
    node: NetNode, lattice: Lattice, collection: Dict[str, Tuple[NetNode, int]]
) -> Tuple[bool, np.ndarray, int]:
    """Add a node to a collection.

    Return a tuple of (is_new, image, index).
    We find image using the call .periodic_image(node, lattice).

    Args:
        node (NetNode): The node to add to the collection.
        lattice (Lattice): The lattice to use for finding the periodic image.
        collection (Dict[str, Tuple[NetNode, int]]): The collection to add the node to.]=

    Returns:
        Tuple[bool, np.ndarray, int]: A tuple of (is_new, image, index).
    """
    image = None
    for n, i in collection.values():
        if node.is_periodic_image(n, lattice):
            image = node.periodic_image(n, lattice)
            return False, image, i

    is_new = True
    collection[str(node)] = (node, len(collection))

    return is_new, image, collection[str(node)][1]


def contains_edge(edge: VoltageEdge, collection: Iterable[VoltageEdge]) -> bool:
    """Return True if the edge is in the collection."""
    return any(e == edge for e in collection)


def branching_index_match(linker, metal_cluster, lattice: Lattice, tolerance: float = 1e-8):
    """Check if the branching index of the linker is the same as of the metal cluster."""
    for branching_coordinate in metal_cluster.branching_coords:

        frac_a = lattice.get_fractional_coords(branching_coordinate)
        for coord in linker.branching_coords:
            frac_b = lattice.get_fractional_coords(coord)
            # wrap the coordinates to the unit cell
            # frac_b = frac_b- np.floor(frac_b)
            distance, image = lattice.get_distance_and_image(frac_a, frac_b)

            if distance < tolerance:
                return True, image
    return False, None


def which_periodic_image(
    a: np.ndarray, b: np.ndarray, lattice: Lattice, tolerance: float = 1e-2, is_frac: bool = True
) -> Tuple[bool, Optional[np.array], Optional[List[Tuple]]]:
    """
    Find which periodic image of a is closest to b.

    Uses pymatgens get_distance_and_image.

    Args:
        a (np.ndarray): The first vector.
        b (np.ndarray): The second vector.
        lattice (Lattice): The lattice to use for finding the periodic image.
        tolerance (float): The tolerance for finding the periodic image.
        is_frac (bool): Whether the vectors are in fractional coordinates.

    Returns:
        Tuple[bool, Optional[np.array], Optional[List[Tuple]]]: A tuple of (is_new, image, image_list).
    """
    if not is_frac:
        a = lattice.get_fractional_coords(a)
        b = lattice.get_fractional_coords(b)
    if is_periodic_image(a, b, lattice, tolerance, is_frac=True):
        _, image = lattice.get_distance_and_image(a, b)
        return True, a, [(np.array([0, 0, 0]), -image), (image, np.array([0, 0, 0]))]
    return False, None, None


def in_cell(node: "SBU", lattice: Lattice) -> Tuple[bool, List[np.array]]:
    """
    Check if a node is in the unit cell.

    It does so by checking if any branching coordinate is in the unit cell.

    Args:
        node (SBU): The node to check.
        lattice (Lattice): The lattice to use for checking.

    Returns:
        Tuple[bool, List[np.array]]: A tuple of (is_in_cell, coordinates in cell).
    """
    branching_coords_in_cell = []
    for branching_coord in node.branching_coords:
        branching_coord = lattice.get_fractional_coords(branching_coord)
        if np.all(branching_coord < 1) & np.all(branching_coord > 0):
            branching_coords_in_cell.append(branching_coord)
    if len(branching_coords_in_cell) == 0:
        return False, branching_coords_in_cell
    return True, branching_coords_in_cell


def com_in_cell(node: "SBU", lattice: Lattice):
    com = node.molecule.center_of_mass
    com_frac = lattice.get_fractional_coords(com)
    if np.all(com_frac < 1) & np.all(com_frac >= 0):
        return True
    return False


def has_edge(
    metal_cluster: NetNode, linker: NetNode, lattice: Lattice
) -> Tuple[bool, List[np.array]]:
    """Check if the linker has an edge to the metal cluster.

    It does so by checking if there is any connecting between the branching indices
    - either in the cell or between periodic replica.

    Args:
        metal_cluster (NetNode): The metal cluster to check.
        linker (NetNode): The linker to check.
        lattice (Lattice): The lattice to use for checking.

    Returns:
        Tuple[bool, List[np.array]]: A tuple of (is_connected, image).
    """
    images = []

    for metal_coord in metal_cluster._all_branching_coord:
        for linker_coord in linker._all_branching_coord:
            match, coord, images_ = which_periodic_image(
                metal_coord, linker_coord, lattice, is_frac=False
            )
            if match:
                for image_a, image_b in images_[:1]:
                    images.append((coord, image_a, image_b))

    return len(images) > 0, images


def build_net(
    metal_clusters: NodeCollection,
    linkers: LinkerCollection,
    lattice: Lattice,
) -> Net:
    """Given a metal cluster and linker collection from the fragmentation, build a net.

    Args:
        metal_clusters (NodeCollection): The metal clusters from the fragmentation.
        linkers (LinkerCollection): The linkers from the fragmentation.
        lattice (Lattice): The lattice of the MOF (e.g. MOF.lattice).

    Returns:
        Net: An object representing the combinatorial net.
    """
    found_metal_nodes = OrderedDict()
    found_linker_nodes = OrderedDict()
    found_edges = []

    for metal_cluster in metal_clusters:

        found_in_cell, coordinates = in_cell(metal_cluster, lattice)
        # if not found_in_cell:
        #     continue
        # there seems to be a mismatch between branching coords and the centers in some cases.
        # ToDo: fix this at the source of the problem.
        center = lattice.get_fractional_coords(metal_cluster.center)
        center = center - np.floor(center)
        center = lattice.get_cartesian_coords(center)
        net_node = NetNode(
            center,
            molecule=metal_cluster,
            branching_coord_in_cell=coordinates,
            all_branching_coord=metal_cluster.branching_coords,
            flavor="metalcluster",
        )

        _, _, metal_index = add_node_to_collection(net_node, lattice, found_metal_nodes)

    for linker in linkers:

        found_in_cell, linker_coordinates = in_cell(linker, lattice)
        # if not found_in_cell:
        #    continue
        center = lattice.get_fractional_coords(linker.center)
        center = center - np.floor(center)
        center = lattice.get_cartesian_coords(center)
        net_node = NetNode(
            center,
            molecule=linker,
            branching_coord_in_cell=linker_coordinates,
            all_branching_coord=linker.branching_coords,
            flavor="linker",
        )
        _, _, metal_index = add_node_to_collection(net_node, lattice, found_linker_nodes)

    linker_index_offset = len(found_metal_nodes)

    logger.debug(f"Found {len(found_metal_nodes)} metal nodes out of {len(metal_clusters)}")
    logger.debug(f"Found {len(found_linker_nodes)} linker nodes out of {len(linkers)}")

    egde_candiates = defaultdict(list)
    for metal_node, metal_index in found_metal_nodes.values():
        for linker_node, linker_index in found_linker_nodes.values():
            at_least_one_edge, images = has_edge(metal_node, linker_node, lattice)
            if at_least_one_edge:
                for coord, image_a, image_b in images:
                    metal_center = lattice.get_fractional_coords(metal_node._coords) + image_a
                    linker_center = lattice.get_fractional_coords(linker_node._coords) + image_b
                    edge = VoltageEdge(
                        linker_center - metal_center,
                        metal_index,
                        linker_index + linker_index_offset,
                        image_a,
                        image_b,
                    )
                    egde_candiates[
                        (round(coord[0], 2), round(coord[1], 2), round(coord[2], 2))
                    ].append((edge, np.abs(image_b).sum()))

    for linker_node, linker_index in found_linker_nodes.values():
        for metal_node, metal_index in found_metal_nodes.values():
            at_least_one_edge, images = has_edge(linker_node, metal_node, lattice)
            if at_least_one_edge:
                for coord, image_a, image_b in images:
                    metal_center = lattice.get_fractional_coords(metal_node._coords) + image_b
                    linker_center = lattice.get_fractional_coords(linker_node._coords) + image_a
                    edge = VoltageEdge(
                        linker_center - metal_center,
                        metal_index,
                        linker_index + linker_index_offset,
                        image_a,
                        image_b,
                    )
                    egde_candiates[
                        (round(coord[0], 2), round(coord[1], 2), round(coord[2], 2))
                    ].append((edge, np.abs(image_b).sum()))

    edge_selection = []
    for _branching_coord, edges in egde_candiates.items():
        # branching_coord = np.array(branching_coord)
        # if np.all(branching_coord < 1) & np.all(
        #     branching_coord > 0
        # ):
        if len(edges) == 1:
            edge_selection.append(edges[0][0])
        else:
            # sort ascending by second element in tuple in the list
            edge_selection.append(sorted(edges, key=lambda x: x[1])[0][0])

    for edge in edge_selection:
        if not contains_edge(edge, found_edges):
            found_edges.append(edge)

    found_metal_nodes.update(found_linker_nodes)
    return Net(found_metal_nodes, found_edges, lattice)


def _simplify_structure_graph(structure_graph: StructureGraph) -> StructureGraph:
    """Simplifies a structure graph by removing two-connected nodes.

    We will place an edge between the nodes that were connected
    by the two-connected node.

    The function does not touch the input graph (it creates a deep copy).
    Using the deep copy simplifies the implementation a lot as we can
    add the edges in a first loop where we check for two-connected nodes
    and then remove the nodes. This avoids the need for dealing with indices
    that might change when one creates a new graph.

    Args:
        structure_graph (StructureGraph): Input structure graph.
            Usually this is a "net graph". That is, a structure graph for
            a structure in which the atoms are MOF SBUs

    Returns:
        StructureGraph: simplified structure graph, where we removed the
            two-connected nodes.
    """
    graph_copy = structure_graph.__copy__()
    to_remove = []
    added_edges = set()

    # in the first iteration we just add the edge
    # and collect the nodes to delete
    for i, _ in enumerate(structure_graph.structure):
        if structure_graph.get_coordination_of_site(i) == 2:
            if str(structure_graph.structure[i].specie) != "Si":
                indices = []
                images = []
                for neighbor in structure_graph.get_connected_sites(i):
                    indices.append(neighbor.index)
                    images.append(neighbor.jimage)
                    try:
                        graph_copy.break_edge(i, neighbor.index, neighbor.jimage)
                    except ValueError:
                        try:
                            graph_copy.break_edge(
                                neighbor.index,
                                i,
                                (-neighbor.jimage[0], -neighbor.jimage[1], -neighbor.jimage[2]),
                            )
                        except ValueError:
                            logger.warning(f"Edge {i, neighbor.index} cannot be broken")
                sorted_images = [x for _, x in sorted(zip(indices, images))]
                edge_tuple = (tuple(sorted(indices)), tuple(sorted_images))
                # in principle, this check should not be needed ...
                if edge_tuple not in added_edges:
                    added_edges.add(edge_tuple)
                    graph_copy.add_edge(indices[0], indices[1], images[0], images[1])

                to_remove.append(i)
            else:
                logger.warning(
                    "Metal cluster with low coodination number detected.\
                    Results might be incorrect."
                )

    # after we added all the edges, we can remove the nodes
    graph_copy.remove_nodes(to_remove)
    return graph_copy



"""Test SBU functionality."""
import pytest


def test_sbu(get_linker_object, get_node_object):
    linker = get_linker_object
    assert linker.composition == "C9 H3 O6"
    assert pytest.approx(linker.molar_mass) == 207.11651999999998

    node = get_node_object
    assert node.composition == "C4 Cu2 O8"
    assert pytest.approx(node.molar_mass) == 303.13



# -*- coding: utf-8 -*-
"""Test locator modules"""
from collections import Counter

from pymatgen.core import Structure

from moffragmentor import MOF
from moffragmentor.fragmentor.linkerlocator import _create_linkers_from_node_location_result
from moffragmentor.fragmentor.nodelocator import create_node_collection, find_node_clusters
from moffragmentor.fragmentor.solventlocator import (
    _get_solvent_molecules_bound_to_node,
    get_all_bound_solvent_molecules,
)
from moffragmentor.molecule import NonSbuMoleculeCollection


def test_find_cu_i_ii_btc_clusters(get_cuiiibtc_mof):
    """Make sure that we do find pytest he two correct node types in Cu(I/II)-BTC"""
    mof = get_cuiiibtc_mof
    node_location_result = find_node_clusters(mof)
    for node in node_location_result.nodes:
        # Every node has four branching indices from the carboxy
        assert len(node & node_location_result.branching_indices) == 4
    assert len(node_location_result) == 5
    assert len(node_location_result.nodes) == 6  # fou r paddlewheels and two macrocylus
    nodes_sizes = [len(node) for node in node_location_result.nodes]
    assert len(set(nodes_sizes)) == 2
    size_counters = Counter(nodes_sizes)
    assert sorted(list(size_counters.values())) == [2, 4]
    assert sorted(list(size_counters.keys())) == [16, 20]


def test_find_porphyrin_mof_clusters(get_porphryin_mof):
    """Looking for the porphyrin MOF clusters"""
    mof = get_porphryin_mof
    node_location_result = find_node_clusters(mof)
    fragments = mof.fragment()
    assert len(node_location_result) == 5
    assert len(node_location_result.nodes) == 4
    assert len(fragments.nodes) == 4  # we by default break the rod nodes
    assert len(node_location_result.branching_indices) == 8 * 2
    assert fragments.linkers[0].search_pubchem()[0][0].cid == 58107362


def test_find_hypothetical_mof_clusters(get_hypothetical_mof):
    """Check if we can find the correct nodes in hypothetical MOFs"""
    mof = get_hypothetical_mof
    node_location_result = find_node_clusters(mof)
    assert len(node_location_result) == 5
    assert len(node_location_result.nodes) == 16
    node_lengths = [len(node) for node in node_location_result.nodes]
    assert len(set(node_lengths)) == 1
    assert node_lengths[0] == 26


def test_find_p_linker_floating_mof_clusters(get_p_linker_with_floating):
    """Check if we can deal with non-carboxy ligand and floating solvent"""
    mof = get_p_linker_with_floating
    node_location_result = find_node_clusters(mof)
    assert len(node_location_result) == 5
    assert len(node_location_result.nodes) == 4
    node_lengths = [len(node) for node in node_location_result.nodes]
    assert len(set(node_lengths)) == 1
    assert node_lengths[0] == 23


def test_find_li_mof_floating_mof_cluster(get_li_mof_with_floating):
    """A somewhat more complicated node,
    see https://pubs.rsc.org/en/content/articlelanding/2014/DT/c3dt53415d#!divAbstract"""
    mof = get_li_mof_with_floating
    node_location_result = find_node_clusters(mof)
    assert len(node_location_result) == 5
    assert len(node_location_result.nodes) == 2
    node_lengths = [len(node) for node in node_location_result.nodes]
    assert len(set(node_lengths)) == 1
    assert node_lengths[0] == 34


def test_find_rod_node_floating_mof_cluster(get_1d_node_with_floating):
    mof = get_1d_node_with_floating
    node_location_result = find_node_clusters(mof)
    assert len(node_location_result) == 5
    assert len(node_location_result.nodes) == 4
    node_lengths = [len(node) for node in node_location_result.nodes]
    assert len(set(node_lengths)) == 1
    assert node_lengths[0] == 20


def test_formate_mof(get_formate_structure_and_graph):
    s, sg = get_formate_structure_and_graph
    mof = MOF(s, sg)
    bbs = mof.fragment()
    assert len(bbs.nodes) == 2
    assert (
        len(bbs.capping_molecules) == 0
    )  # since there is no linker, we assign the formate as capping molecule
    assert bbs.linkers.composition == {"C1 H1 O2": 6}


def test__get_solvent_molecules_bound_to_node(get_li_mof_with_floating):
    mof = get_li_mof_with_floating
    node_indices = [
        0,
        128,
        130,
        2,
        132,
        4,
        134,
        6,
        136,
        138,
        140,
        142,
        144,
        146,
        148,
        150,
        152,
        154,
        156,
        158,
        164,
        166,
        48,
        50,
        52,
        54,
        68,
        70,
        92,
        94,
        104,
        106,
        116,
        118,
    ]
    solvent_molecules = _get_solvent_molecules_bound_to_node(mof, node_indices)
    assert isinstance(solvent_molecules, NonSbuMoleculeCollection)
    assert solvent_molecules.composition == {"H2 O1": 2}


def test_find_all_bound_solvent_molecules(get_li_mof_with_floating):
    mof = get_li_mof_with_floating
    node_collection = [
        {
            0,
            128,
            130,
            2,
            132,
            4,
            134,
            6,
            136,
            138,
            140,
            142,
            144,
            146,
            148,
            150,
            152,
            154,
            156,
            158,
            164,
            166,
            48,
            50,
            52,
            54,
            68,
            70,
            92,
            94,
            104,
            106,
            116,
            118,
        },
        {
            1,
            129,
            3,
            131,
            133,
            5,
            135,
            7,
            137,
            139,
            141,
            143,
            145,
            147,
            149,
            151,
            153,
            155,
            157,
            159,
            165,
            167,
            49,
            51,
            53,
            55,
            69,
            71,
            93,
            95,
            105,
            107,
            117,
            119,
        },
    ]
    solvent_molecules = get_all_bound_solvent_molecules(mof, node_collection)
    assert isinstance(solvent_molecules, NonSbuMoleculeCollection)
    assert solvent_molecules.composition == {"H2 O1": 4}


def test_find_node_cluster_acetate_zr_mof(get_acetate_zr_mof):
    """Nice (Zr6)2 node,
    see https://pubs.rsc.org/en/content/articlelanding/2018/CC/C8CC00507A#!divAbstract"""
    mof = get_acetate_zr_mof
    node_location_result = find_node_clusters(mof)
    assert len(node_location_result) == 5
    assert len(node_location_result.nodes) == 1
    # there are only three carboxy that actually come from a linker in Zr6 part
    assert len(node_location_result.branching_indices) == 24
    node_lengths = [len(node) for node in node_location_result.nodes]
    assert len(set(node_lengths)) == 1


def test_find_all_bound_solvent_molecules_acetate_zr_mof(get_acetate_zr_mof):
    mof = get_acetate_zr_mof
    node_collection = [
        {
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            118,
            119,
            122,
            123,
            126,
            127,
            130,
            131,
            134,
            135,
            138,
            139,
            142,
            143,
            144,
            145,
            170,
            171,
            184,
            185,
            188,
            189,
            192,
            193,
            202,
            203,
            204,
            205,
            206,
            207,
            208,
            209,
            210,
            211,
            212,
            213,
            214,
            215,
            216,
            217,
            218,
            219,
            220,
            221,
            222,
            223,
            224,
            225,
            226,
            227,
            228,
            229,
            230,
            231,
            232,
            233,
            234,
            235,
            236,
            237,
            238,
            239,
            240,
            241,
            242,
            243,
            244,
            245,
            246,
            247,
            248,
            249,
            250,
            251,
            252,
            253,
            254,
            255,
            256,
            257,
            258,
            259,
            260,
            261,
            262,
            263,
            264,
            265,
        }
    ]
    solvent_molecules = get_all_bound_solvent_molecules(mof, node_collection)
    assert isinstance(solvent_molecules, NonSbuMoleculeCollection)
    assert solvent_molecules.composition == {"C2 H3 O2": 8}


def test__create_linkers_from_node_location_result(get_hkust_mof):
    mof = get_hkust_mof
    unbound_solvent = NonSbuMoleculeCollection([])
    node_location_result = find_node_clusters(mof)
    node_collection = create_node_collection(mof, node_location_result)
    linkers = _create_linkers_from_node_location_result(
        mof, node_location_result, node_collection, unbound_solvent, unbound_solvent
    )
    assert len(linkers) == 8
    linker_lengths = [len(linker) for linker in linkers]
    assert len(set(linker_lengths)) == 1


def test__get_node_cluster_across_pbc(get_across_periodic_boundary_node):
    mof = get_across_periodic_boundary_node
    node_location_result = find_node_clusters(mof)
    assert len(node_location_result.nodes) == 2  # we have 8 Zn in unit cell and Zn4 nodes
    sites = []

    for site in node_location_result.nodes[0]:
        sites.append(mof.structure[site])
    s = Structure.from_sites(sites)
    assert str(s.composition) == "Zn4 P2 H6 C6 O18"



# -*- coding: utf-8 -*-
"""Errors reused across the moffragmentor package"""
__all__ = ["JavaNotFoundError", "NoMetalError"]


class JavaNotFoundError(ValueError):
    """Raised if Java executable could not be found"""


class NoMetalError(ValueError):
    """Raised if structure contains no metal"""



# -*- coding: utf-8 -*-
"""Generate molecules as the subgraphs from graphs"""
from collections import defaultdict
from typing import Iterable, Optional

import numpy as np
from loguru import logger
from pymatgen.core import Molecule, Site
from structuregraph_helpers.create import VestaCutoffDictNN

from .. import mof


def wrap_molecule(
    mol_idxs: Iterable[int],
    mof: "mof.MOF",
    starting_index: Optional[int] = None,
    add_additional_site: bool = True,  # noqa: F821
) -> Molecule:
    """Wrap a molecule in the cell of the MOF by walking along the structure graph.

    For this we perform BFS from the starting index. That is, we use a queue to
    keep track of the indices of the atoms that we still need to visit
    (the neighbors of the current index).
    We then compute new coordinates by computing the Cartesian coordinates
    of the neighbor image closest to the new coordinates of the current atom.

    To then create a Molecule with the correct ordering of sites, we walk
    through the hash table in the order of the original indices.

    Args:
        mol_idxs (Iterable[int]): The indices of the atoms in the molecule in the MOF.
        mof (MOF): MOF object that contains the mol_idxs.
        starting_index (int, optional): Starting index for the walk.
            Defaults to 0.
        add_additional_site (bool): Whether to add an additional site

    Returns:
        Molecule: wrapped molecule
    """
    if starting_index is None:
        if len(mol_idxs) == 1:
            starting_index = 0
        # take the index of the atom which coordinates are closest to the origin
        else:
            # Here was a bug before because i missed the zip
            starting_index = min(
                (zip(np.arange(len(mol_idxs)), mof.structure.cart_coords[mol_idxs])),
                key=lambda x: np.linalg.norm(x[1]),
            )[0]

    new_positions_cart = {}
    new_positions_frac = {}
    still_to_wrap_queue = [mol_idxs[starting_index]]
    new_positions_cart[mol_idxs[starting_index]] = mof.cart_coords[mol_idxs[starting_index]]
    new_positions_frac[mol_idxs[starting_index]] = mof.frac_coords[mol_idxs[starting_index]]
    additional_sites = []
    while still_to_wrap_queue:
        current_index = still_to_wrap_queue.pop(0)
        if current_index in mol_idxs:
            neighbor_indices = mof.get_neighbor_indices(current_index)
            for neighbor_index in neighbor_indices:
                if (neighbor_index not in new_positions_cart) & (neighbor_index in mol_idxs):
                    _, image = mof.structure[neighbor_index].distance_and_image_from_frac_coords(
                        new_positions_frac[current_index]
                    )
                    new_positions_frac[neighbor_index] = mof.frac_coords[neighbor_index] - image
                    new_positions_cart[neighbor_index] = mof.lattice.get_cartesian_coords(
                        new_positions_frac[neighbor_index]
                    )
                    still_to_wrap_queue.append(neighbor_index)
                else:
                    if neighbor_index in new_positions_cart:
                        species_a = str(mof.structure[current_index].specie)
                        species_b = str(mof.structure[neighbor_index].specie)

                        if (
                            np.linalg.norm(
                                new_positions_cart[neighbor_index]
                                - new_positions_cart[current_index]
                            )
                            > VestaCutoffDictNN._lookup_dict[species_a][species_b]
                        ) & add_additional_site:
                            logger.warning(
                                "Warning: neighbor_index {} is already in new_positions_cart, "
                                "but the distance is too large. "
                                "Will add an additional site. This is unusual and not well tested".format(
                                    neighbor_index
                                )
                            )

                            _, image = mof.structure[
                                neighbor_index
                            ].distance_and_image_from_frac_coords(new_positions_frac[current_index])
                            new_frac = mof.frac_coords[neighbor_index] - image
                            new_cart = mof.lattice.get_cartesian_coords(new_frac)
                            additional_sites.append((neighbor_index, new_cart, new_frac))

    new_sites = []
    for _, idx in enumerate(mol_idxs):
        new_sites.append(Site(mof.structure[idx].species, new_positions_cart[idx]))

    idx_mapping = defaultdict(list)
    for i, idx in enumerate(mol_idxs):
        idx_mapping[i].append(idx)
    for i, (idx, cart, _) in enumerate(additional_sites):
        new_sites.append(Site(mof.structure[idx].species, cart))
        idx_mapping[i + len(idx_mapping)].append(idx)

    return Molecule.from_sites(new_sites), idx_mapping



"""Test SBU collection functionality."""
from moffragmentor.sbu.sbucollection import SBUCollection


def test_sbu_collection(get_linker_object):
    sbu_collection = SBUCollection([get_linker_object])
    assert len(sbu_collection.sbus) == 1
    assert sbu_collection.sbus[0].composition == "C9 H3 O6"

    sbu_collection = SBUCollection([get_linker_object, get_linker_object])
    assert len(sbu_collection.sbus) == 2



from pymatgen.core import Structure
from structuregraph_helpers.subgraph import get_subgraphs_as_molecules

from moffragmentor.fragmentor.branching_points import has_metal_in_path
from moffragmentor.fragmentor.linkerlocator import _pick_central_linker_indices
from moffragmentor.fragmentor.molfromgraph import wrap_molecule
from moffragmentor.sbu.node import Node
from moffragmentor.sbu.nodecollection import NodeCollection


def generate_new_node_collection(mof, node_result):

    graph_ = mof.structure_graph.__copy__()
    graph_.structure = Structure.from_sites(graph_.structure.sites)
    graph_.remove_nodes(node_result.branching_indices)

    mols, graphs, idxs, centers, coordinates = get_subgraphs_as_molecules(
        graph_,
        return_unique=False,
        filter_in_cell=False,
        disable_boundary_crossing_check=True,
        prune_long_edges=True,
    )
    node_indices, coords = _pick_central_linker_indices(mof, coordinates)
    node_indices = [i for i in node_indices if has_metal_in_path(mof, idxs[i])]
    mols = [mol for i, mol in enumerate(mols) if i in node_indices]

    branching_sites_for_node = {}

    for i in node_indices:
        neighbor_pool = set()
        for vertex in idxs[i]:
            neighbor_pool.update(mof.get_neighbor_indices(vertex))

        branching_sites_for_node[i] = neighbor_pool & node_result.branching_indices

    nodes = []
    found_hashes = set()
    for i, node_index in enumerate(node_indices):
        idx = idxs[node_index]
        center = centers[node_index]
        coords_ = coords[i]
        branching_indices = list(branching_sites_for_node[node_index])
        mol, mapping = wrap_molecule(idx + branching_indices, mof)

        node = Node(
            molecule=mol,
            molecule_graph=graphs[node_index],
            center=center,
            graph_branching_indices=branching_indices,
            closest_branching_index_in_molecule=branching_indices,
            binding_indices=[],
            coordinates=coords_,
            original_indices=idx + list(branching_indices),
            connecting_paths=[],
            molecule_original_indices_mapping=mapping,
        )
        if node.hash not in found_hashes:
            found_hashes.add(node.hash)
            nodes.append(node)

    return NodeCollection(nodes)



# -*- coding: utf-8 -*-
"""Describing an collection of linkers"""
from typing import List

from pymatgen.core import Molecule

from .sbucollection import SBUCollection

__all__ = ["LinkerCollection"]


class LinkerCollection(SBUCollection):
    """Collection of linker building blocks"""

    @property
    def building_block_composition(self) -> List[str]:
        """Return a list of strings of building blocks.

        Strings are of the form of L{i} where i is an integer.

        Returns:
            List[str]: List of strings of building blocks.
        """
        if self._composition is None:
            self._composition = ["".join(["L", i]) for i in self.sbu_types]
        return self._composition

    def _linker_structure(self):
        all_linker_sites = []

        for n in self:
            n.molecule.coords = n.cart_coords
            s = n.molecule.sites
            all_linker_sites.extend(s)

        super_node = Molecule.from_sites(all_linker_sites)
        return super_node

    def show_linker_structure(self):
        import nglview as nv

        super_node = self._linker_structure()
        return nv.show_pymatgen(super_node)



# -*- coding: utf-8 -*-
"""This module contains functions that perform filtering on indices or fragments.

Those fragments are typically obtained from the other fragmentation modules.
"""

from typing import Iterable, Union

import networkx as nx
import numpy as np
from loguru import logger
from pymatgen.core import Lattice
from scipy.spatial.qhull import Delaunay, QhullError  # pylint:disable=no-name-in-module

from .. import mof
from ..utils import unwrap


def bridges_across_cell(mof: "mof.MOF", indices: Iterable[int]) -> bool:
    """Check if a molecule of indices bridges across the cell"""
    bridges = {}

    for index in indices:
        for neighbor_site in mof.structure_graph.get_connected_sites(index):
            if neighbor_site.index not in bridges:
                bridges[neighbor_site.index] = neighbor_site.jimage
            else:
                if (bridges[neighbor_site.index] != neighbor_site.jimage) & (
                    neighbor_site.index in mof.metal_indices
                ):
                    return True

    return False


def point_in_mol_coords(point: np.array, points: np.array, lattice: Lattice) -> bool:
    # perhaps rather do via the COM
    new_coords = unwrap(points, lattice)
    new_point = unwrap(point, lattice)

    return (
        in_hull(point, new_coords)
        or in_hull(point, points)
        or in_hull(new_point, points)
        or in_hull(new_point, new_coords)
    )


def in_hull(pointcloud: np.array, hull: Union[np.array, Delaunay]) -> bool:
    """
    Test if points in `p` are in `hull`.

    Taken from https://stackoverflow.com/a/16898636

    Args:
        pointcloud (np.array): points to test (`NxK` coordinates of `N` points in `K` dimensions)
        hull (np.array): Is either a scipy.spatial.Delaunay object
            or the `MxK` array of the coordinates of `M` points in `K` dimensions
            for which Delaunay triangulation will be computed

    Returns:
        bool: True if all points are in the hull, False otherwise
    """
    try:
        if not isinstance(hull, Delaunay):
            hull = Delaunay(hull)
    except (QhullError, ValueError):

        if len(pointcloud) < 5:
            logger.warning("Too few points to compute Delaunay triangulation")
            return False

        hull = Delaunay(hull, qhull_options="QJ")

    return hull.find_simplex(pointcloud) >= 0


def _filter_branch_points(branch_indices: list, metal_indices: list, graph: nx.Graph) -> list:
    """Filter incorrectly identified branch points.

    In a MOF structure there might be many sites with
    more than three neighbors that do not lead to a tree or
    leaf node. The relevant branching indices are those that
    are not between other ones.

    That is, we want to filter out branching indices for which the shortest
    path to a metal goes via another branching index.

    Args:
        branch_indices (list): candidate list of branching indices
        metal_indices (list): metal indices
        graph (nx.Graph): graph on which the nodes can be access using the
            items on the branch_indices and metal_indices lists

    Returns:
        list filtered branching indices
    """
    filtered_indices = []
    for branch_index in branch_indices:
        shortest_path = _shortest_path_to_metal(branch_index, metal_indices, graph)
        if not _has_branch_index_in_path(shortest_path, branch_indices):
            filtered_indices.append(branch_index)
    return filtered_indices


def _has_branch_index_in_path(path, branch_indices):
    return any(metal_index in path for metal_index in branch_indices)


def _shortest_path_to_metal(branch_index, metal_indices, graph):
    paths = []

    for metal_index in metal_indices:
        path = nx.shortest_path(  # pylint:disable=unexpected-keyword-arg, no-value-for-parameter
            graph, source=branch_index, target=metal_index
        )
        paths.append(path)

    shortest_path = min(paths, key=len)

    return shortest_path[1:]



# -*- coding: utf-8 -*-
# pylint: disable=line-too-long
"""Methods for running systre"""
import os
import re
import subprocess
from collections import defaultdict
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
from typing import List, Tuple

from loguru import logger
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.core import Lattice

from . import is_tool
from .errors import JavaNotFoundError

__all__ = ["run_systre"]

THIS_DIR = os.path.dirname(os.path.realpath(__file__))

SYSTRE_JAR = os.path.abspath(os.path.join(THIS_DIR, "..", "Systre-19.6.0.jar"))


@contextmanager
def closed_named_tempfile(contents, mode="w", suffix=".cdg"):
    file = NamedTemporaryFile(delete=False, mode=mode, suffix=suffix)
    try:
        with file:
            file.write(contents)
        yield file.name
    finally:
        os.unlink(file.name)


def run_systre(systre_string: str) -> dict:
    if not is_tool("java"):
        raise JavaNotFoundError("To determine the topology of the net, `java` must be in the PATH.")
    try:
        with closed_named_tempfile(systre_string, suffix=".cgd", mode="w") as filename:
            cmd_list = [
                "java",
                "-cp",
                str(SYSTRE_JAR),
                "org.gavrog.apps.systre.SystreCmdline",
                filename,
            ]
            out = subprocess.run(
                cmd_list,
                universal_newlines=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
            systre_result = _parse_systre_lines(out.stdout.split("\n"))

            return systre_result
    except Exception:  # pylint:disable=broad-except
        logger.warning("Error running Systre.")
        return ""


def _parse_systre_lines(
    lines: List[str],
) -> dict:
    """Parse the output of Systre.

    Input for this function might be created with

    ```python
    with open('systre.out', 'r') as handle:
        lines = handle.readlines()
    ```

    Args:
        lines (List[str]): Output of systre run,
            parsed into a list of strings (one line per string)

    Returns:
        dict: parsed output
    """
    rcsr_line = float("inf")
    cell_line = float("inf")
    angle_line = float("inf")
    in_node_block = False
    in_edge_block = False
    node_start_line = float("inf")
    edge_start_line = float("inf")
    nodes = defaultdict(list)
    edges = []
    for i, line in enumerate(lines):
        if "Ideal space group" in line:
            space_group = line.split("is")[-1].strip().replace(".", "")
        if "Structure was identified with RCSR symbol:" in line:
            rcsr_line = i + 1
        if i == rcsr_line:
            rcsr_code = line.split()[-1].strip()
        if "Relaxed cell parameters" in line:
            cell_line = i + 1
            angle_line = i + 2
        if i == cell_line:
            cell = _line_to_coords(line)
        if i == angle_line:
            angles = _line_to_coords(line)
        if "Relaxed positions:" in line:
            node_start_line = i + 1
        if i == node_start_line:
            in_node_block = True
        if i == edge_start_line:
            in_edge_block = True
        if "Edges:" in line:
            in_node_block = False
            edge_start_line = i + 1
        if "Edge centers:" in line:
            in_edge_block = False
        if in_node_block:
            try:
                node, coords = _parse_node_line(line)
                nodes[node].append(coords)
            except IndexError:
                pass

        if in_edge_block:
            try:
                edge_coords = _line_to_coords(line)
                edges.append((edge_coords[:3], edge_coords[3:]))
            except IndexError:
                pass

    results = {
        "space_group": space_group,
        "rcsr_code": rcsr_code,
        "relaxed_cell": cell,
        "relaxed_angles": angles,
        "relaxed_node_positions": dict(nodes),
        "relaxed_edges": edges,
    }

    return results


def _line_to_coords(line: str) -> List[float]:
    return [float(i) for i in re.findall(r"\d+.\d+", line)]


def _parse_node_line(line: str) -> Tuple[int, List[float]]:
    node_number = int(re.findall(r"Node\s(\d+):", line)[0])
    coords = _line_to_coords(line)

    return node_number, coords


def _get_systre_input_from_pmg_structure_graph(
    structure_graph: StructureGraph, lattice: Lattice = None
) -> str:
    """
    Loop over all atoms in a StructureGraph and use them as nodes.

    Place edges such that all nodes are represented with their
    full connectivity.

    Args:
        structure_graph (StructureGraph): pymatgen StructureGraph
            object representing the net. This will correspond to having
            one atom per SBU in the structure graph.
        lattice (Lattice): pymatgen lattice object. Needed for the cell dimension

    Returns:
        str: systre input string. Does not contain the optional edge centers
    """
    lattice = structure_graph.structure.lattice if lattice is None else lattice
    vertices = []
    edges = []

    symmetry_group = "   GROUP P1"

    cell_line = f"   CELL {lattice.a} {lattice.b} {lattice.c} {lattice.alpha} {lattice.beta} {lattice.gamma}"

    frac_coords = structure_graph.structure.frac_coords

    for i in range(len(structure_graph)):
        vertices.append((len(structure_graph.get_connected_sites(i)), frac_coords[i]))
        # vertices.append((structure_graph.get_coordination_of_site(i), frac_coords[i]))

    for edge in structure_graph.graph.edges(data=True):
        start = frac_coords[edge[0]]
        end = frac_coords[edge[1]] + edge[2]["to_jimage"]
        edges.append((tuple(start), tuple(end)))

    def _create_vertex_string(counter, coordination, coordinate):
        return f"   NODE {counter} {coordination} {coordinate[0]:.4f} {coordinate[1]:.4f} {coordinate[2]:.4f}"

    def _create_edge_string(coordinate_0, coordinate_1):
        return f"   EDGE {coordinate_0[0]:.4f} {coordinate_0[1]:.4f} {coordinate_0[2]:.4f} {coordinate_1[0]:.4f} {coordinate_1[1]:.4f} {coordinate_1[2]:.4f}"  # noqa:E501

    edge_lines = set()
    vertex_lines = set()

    for i, vertex in enumerate(vertices):
        vertex_lines.add(_create_vertex_string(i, vertex[0], vertex[1]))

    for _, edge in enumerate(edges):
        edge_lines.add(_create_edge_string(edge[0], edge[1]))

    file_lines = (
        ["CRYSTAL", "   NAME", symmetry_group, cell_line]
        + list(vertex_lines)
        + list(edge_lines)
        + ["END"]
    )

    return "\n".join(file_lines)



# -*- coding: utf-8 -*-
"""Testing the splitting module of the fragmentor subpackage"""
from collections import Counter

from structuregraph_helpers.subgraph import get_subgraphs_as_molecules

from moffragmentor.fragmentor.solventlocator import get_floating_solvent_molecules


def test_unbound_solvent_identification(get_p_linker_with_floating):
    """Test that we can correctly flag the floating solvents"""
    unique_mols, unique_graphs, unique_indices, _, _ = get_subgraphs_as_molecules(
        get_p_linker_with_floating.structure_graph
    )
    # we see pyrrolidinium, h2o and h3o+ as solvent,
    # see https://www.ccdc.cam.ac.uk/structures/Search?Ccdcid=MAGBON&DatabaseToSearch=Published
    assert len(unique_mols) == len(unique_graphs) == len(unique_indices) == 3
    compositions = [str(unique_mol.composition.alphabetical_formula) for unique_mol in unique_mols]

    assert "H2 O1" in compositions
    assert "C4 H10 N1" in compositions
    assert "O1" in compositions

    # Now, we'll return all in the unit cell
    mols, _, _, _, _ = get_subgraphs_as_molecules(
        get_p_linker_with_floating.structure_graph, return_unique=False
    )
    assert len(mols) == 24
    compositions = [str(mol.composition.alphabetical_formula) for mol in mols]
    composition_counter = Counter(compositions)
    assert composition_counter["C4 H10 N1"] == 4
    assert composition_counter["H2 O1"] == 8 * 2


def test_unbound_solvent_identification_li_mof(get_li_mof_with_floating):
    """Test flagging of unbound solvent in a MOF
    with a Li containing node"""
    unique_mols, unique_graphs, unique_indices, _, _ = get_subgraphs_as_molecules(
        get_li_mof_with_floating.structure_graph
    )
    assert len(unique_mols) == len(unique_graphs) == len(unique_indices) == 1
    compositions = [str(unique_mol.composition.alphabetical_formula) for unique_mol in unique_mols]
    assert "H2 O1" in compositions

    # Now, we'll return all in the unit cell
    mols, graphs, indices, _, _ = get_subgraphs_as_molecules(
        get_li_mof_with_floating.structure_graph, return_unique=False
    )

    assert len(mols) == len(graphs) == len(indices) == 8


def unbound_solvent_identification_li_mof(get_li_mof_with_floating):
    """Test if we can get a NonSbuMoleculeCollection from the MOF"""
    molecules = get_floating_solvent_molecules(get_li_mof_with_floating)
    assert len(molecules) == 8
    assert molecules.composition == {"H2 O1": 8}


def unbound_solvent_identification_acetate_zr_mof(get_acetate_zr_mof):
    molecules = get_floating_solvent_molecules(get_acetate_zr_mof)
    assert len(molecules) == 6
    assert molecules.composition == {"C2 H4 O2": 6}



# -*- coding: utf-8 -*-
"""Test calculation of the SBU dimensionality"""
from moffragmentor.descriptors.sbu_dimensionality import get_structure_graph_dimensionality


def test_get_structure_graph_dimensionality(get_0d_node_graph, get_1d_node_graph):
    """The function we test here does not do anything special
    other than calling pymatgen.
    This test is rather for documentation
    and to check if the function works as a MOF chemist
    would expect"""
    assert get_structure_graph_dimensionality(get_0d_node_graph) == 0
    assert get_structure_graph_dimensionality(get_1d_node_graph) == 1



# -*- coding: utf-8 -*-
"""Dealing with molecules that not part of a secondary building unit."""
from typing import List, Optional

from pymatgen.analysis.graphs import MoleculeGraph, StructureGraph
from pymatgen.core import Molecule

from ..utils import get_edge_dict, remove_all_nodes_not_in_indices


class NonSbuMolecule:
    """Class to handle solvent or other non-SBU molecules."""

    def __init__(
        self,
        molecule: Molecule,
        molecule_graph: MoleculeGraph,
        indices: List[int],
        connecting_index: Optional[int] = None,
    ):
        """Construct a NonSbuMolecule.

        Args:
            molecule (Molecule): A pymatgen Molecule object.
            molecule_graph (MoleculeGraph): A pymatgen MoleculeGraph object.
            indices (List[int]): List of indices of the molecule
                in the original structure.
            connecting_index (int, optional): List of connecting indices
                in the original structure. Defaults to None.
        """
        self.molecule = molecule
        self.molecule_graph = molecule_graph
        self.indices = indices
        # We store the connecting index to see which atom we would
        # need to give the elctron from the bond to the metal,
        # it is not used atm but (hopefully) will be
        self.connecting_index = connecting_index

    @property
    def composition(self) -> str:
        return self.molecule.composition.alphabetical_formula

    def __str__(self):
        """Return string representation of the molecule (the composition)."""
        return str(self.composition)

    def __len__(self):
        """Return number of atoms in the molecule"""
        return len(self.molecule)

    @classmethod
    def from_structure_graph_and_indices(
        cls, structure_graph: StructureGraph, indices: List[int]
    ) -> "NonSbuMolecule":
        """Create a a new NonSbuMolecule from a part of a structure graph.

        Args:
            structure_graph (StructureGraph): Structure graph with structure attribute
            indices (List[int]): Indices that label nodes in the structure graph,
                indexing the molecule of interest

        Returns:
            NonSbuMolecule: Instance of NonSbuMolecule
        """
        my_graph = structure_graph.__copy__()
        remove_all_nodes_not_in_indices(my_graph, indices)
        structure = my_graph.structure
        sites = []
        for site in structure:
            sites.append(site)
        mol = Molecule.from_sites(sites)
        molecule_graph = MoleculeGraph.with_edges(mol, get_edge_dict(my_graph))
        return cls(mol, molecule_graph, indices)

    def show_molecule(self):
        """Use nglview to show the molecule."""
        import nglview  # pylint:disable=import-outside-toplevel

        return nglview.show_pymatgen(self.molecule)



# -*- coding: utf-8 -*-
"""Some pure functions that are used to perform the node identification.

Node classification techniques described
in https://pubs.acs.org/doi/pdf/10.1021/acs.cgd.8b00126.
"""
from collections import namedtuple
from typing import Iterable, List, Optional

import networkx as nx
from loguru import logger
from skspatial.objects import Points

from moffragmentor.descriptors.sbu_dimensionality import get_sbu_dimensionality
from moffragmentor.fragmentor._graphsearch import (
    _complete_graph,
    _to_graph,
    recursive_dfs_until_branch,
    recursive_dfs_until_cn3,
)
from moffragmentor.sbu import Node, NodeCollection
from moffragmentor.utils import _flatten_list_of_sets, _get_metal_sublist

__all__ = [
    "find_node_clusters",
    "create_node_collection",
    "NodelocationResult",
]

NodelocationResult = namedtuple(
    "NodelocationResult",
    [
        "nodes",
        "branching_indices",
        "connecting_paths",
        "binding_indices",
        "to_terminal_from_branching",
    ],
)


def _count_metals_in_path(path, metal_indices):
    """Count the number of metals in a path.

    Args:
        path (List[int]): path to check
        metal_indices (List[int]): indices of metals

    Returns:
        int: number of metals in the path
    """
    return len(set(path) & set(metal_indices))


def _path_without_metal_and_branching_sites(path, metal_indices, branching_indices):
    """Remove the metal indices and the branching indices from a path.

    Args:
        path (List[int]): path to remove the indices from
        metal_indices (List[int]): indices of metals
        branching_indices (List[int]): indices of branching sites

    Returns:
        List[int]: path without metal indices and branching indices
    """
    return [i for i in path if i not in metal_indices and i not in branching_indices]


def find_node_clusters(
    mof,
    unbound_solvent_indices: Optional[List[int]] = None,
    forbidden_indices: Optional[List[int]] = None,
) -> NodelocationResult:
    """Locate the branching indices, and node clusters in MOFs.

    Starting from the metal indices it performs depth first search
    on the structure graph up to branching points.

    Args:
        mof (MOF): moffragmentor MOF instance
        unbound_solvent_indices (List[int], optionl):
            indices of unbound solvent atoms. Defaults to None.
        forbidden_indices (List[int], optional):
            indices not considered as metals, for instance, because
            they are part of a linker. Defaults to None.

    Returns:
        NodelocationResult: nametuple with the slots "nodes", "branching_indices" and
            "connecting_paths"
    """
    logger.debug("Locating node clusters")
    paths = []
    branch_sites = []

    connecting_paths_ = set()

    if forbidden_indices is None:
        forbidden_indices = []

    metal_indices = set([i for i in mof.metal_indices if i not in forbidden_indices])

    if unbound_solvent_indices is None:
        unbound_solvent_indices = []
    # From every metal index in the structure perform DFS up to a
    # branch point
    for metal_index in metal_indices:
        if metal_index not in unbound_solvent_indices:
            p, b = recursive_dfs_until_branch(mof, metal_index, [], [])
            paths.append(p)
            branch_sites.append(b)
    if len(_flatten_list_of_sets(branch_sites)) == 0:
        logger.warning(
            "No branch sites found according to branch site definition.\
             Using now CN=3 sites between metals as branch sites.\
                This is not consistent with the conventions \
                    used in other parts of the code."
        )
        paths = []
        connecting_paths_ = set()
        branch_sites = []

        for metal_index in metal_indices:
            if metal_index not in unbound_solvent_indices:
                p, b = recursive_dfs_until_cn3(mof, metal_index, [], [])
                paths.append(p)
                branch_sites.append(b)

    # The complete_graph will add the "capping sites" like bridging OH
    # or capping formate
    paths, to_terminal_from_branching = _complete_graph(mof, paths, branch_sites)
    # we find the connected components in those paths
    g = _to_graph(mof, paths, branch_sites)
    nodes = list(nx.connected_components(g))

    bs = set(sum(branch_sites, []))

    logger.debug("Locating binding sites and connecting paths")
    binding_sites = set()
    # we store the shortest paths between nodes and branching indices
    # ToDo: we can extract this from the DFS paths above
    for metal, branch_sites_for_metal in zip(metal_indices, branch_sites):
        for branch_site in branch_sites_for_metal:
            # note: using all_simple_paths here is computationally prohibitive
            paths = nx.all_shortest_paths(mof.nx_graph, metal, branch_site)
            for p in paths:
                metals_in_path = _count_metals_in_path(p, metal_indices)
                if metals_in_path >= 1:
                    connecting_paths_.update(p)
                if metals_in_path == 1:
                    binding_path = _path_without_metal_and_branching_sites(p, metal_indices, bs)
                    binding_sites.update(binding_path)
                    # traverse to also add things like Hs in the binding path
                    for site in binding_path:
                        for neighbor in mof.get_neighbor_indices(site):
                            if neighbor not in metal_indices | bs:
                                binding_sites.add(neighbor)

    all_neighbors = []
    for node in nodes:
        neighbors = mof.get_neighbor_indices(node)
        all_neighbors.extend(neighbors)
        intesection = set(neighbors) & bs
        if len(intesection) > 0:
            connecting_paths_.update(neighbors)

    # from the connecting paths we remove the metal indices and the branching indices
    # we need to remove the metal indices as otherwise the fragmentation breaks
    connecting_paths_ -= set(metal_indices)

    res = NodelocationResult(
        nodes, bs, connecting_paths_, binding_sites, to_terminal_from_branching
    )
    return res


def create_node_collection(mof, node_location_result: NodelocationResult) -> NodeCollection:
    # ToDo: This is a bit indirect,
    # it would be better if we would have a list of dicts to loop over
    nodes = []
    for i, _ in enumerate(node_location_result.nodes):
        node_indices = node_location_result.nodes[i]
        node = Node.from_mof_and_indices(
            mof=mof,
            node_indices=node_indices,
            branching_indices=node_location_result.branching_indices,
            binding_indices=node_location_result.binding_indices,
        )
        nodes.append(node)

    return NodeCollection(nodes)


def break_rod_node(mof, indices):
    metal_subset = {i for i in indices if i in mof.metal_indices}
    if not isinstance(metal_subset, int):
        return [{i} for i in metal_subset]
    else:
        return [{metal_subset}]


def create_single_metal_nodes(mof, node_result):
    new_nodes = []
    for node in node_result.nodes:
        new_nodes.extend(break_rod_node(mof, node))
    new_node_result = NodelocationResult(
        new_nodes,
        node_result.branching_indices,
        node_result.connecting_paths,
        node_result.binding_indices,
        node_result.to_terminal_from_branching,
    )
    return new_node_result


def break_rod_nodes(mof, node_result):
    """Break rod nodes into smaller pieces."""
    new_nodes = []
    for node in node_result.nodes:
        if get_sbu_dimensionality(mof, node) >= 1:
            logger.debug("Found 1- or 2-dimensional node. Will break into isolated metals.")
            new_nodes.extend(break_rod_node(mof, node))
        else:
            new_nodes.append(node)
    new_node_result = NodelocationResult(
        new_nodes,
        node_result.branching_indices,
        node_result.connecting_paths,
        node_result.binding_indices,
        node_result.to_terminal_from_branching,
    )
    return new_node_result


def check_node(node_indices, branching_indices, mof) -> bool:
    """Check if the node seems to be a reasonable SBU.

    If not, we can use this to change the fragmentation to something more robust.

    Args:
        node_indices (set): set of indices that make up the node
        branching_indices (set): set of indices that are branching indices
        mof (MOF): MOF object

    Returns:
        bool: True if the node is reasonable, False otherwise
    """
    # check if there is not way more organic than metal
    num_organic = len(node_indices & set(mof.c_indices)) + len(node_indices & set(mof.n_indices))
    num_metals = len(node_indices & set(mof.metal_indices))
    branching_indices_in_node = branching_indices & node_indices
    if num_organic > num_metals + len(branching_indices_in_node):
        return False
    return True


def break_organic_nodes(node_result, mof):
    """If we have a node that is mostly organic, we break it up into smaller pieces."""
    new_nodes = []
    for node in node_result.nodes:
        if len(node) == len(mof):
            logger.debug("Breaking node as full MOF is assigned as node.")
            new_nodes.extend(break_rod_node(mof, node))
        elif check_node(node, node_result.branching_indices, mof) or might_be_porphyrin(
            node, node_result.branching_indices, mof
        ):
            new_nodes.append(node)
        else:
            logger.debug(
                f"Found node {node} that is mostly organic. Will break into isolated metals."
            )
            new_nodes.extend(break_rod_node(mof, node))
    return NodelocationResult(
        new_nodes,
        node_result.branching_indices,
        node_result.connecting_paths,
        node_result.binding_indices,
        node_result.to_terminal_from_branching,
    )


def find_nodes(
    mof,
    unbound_solvent: "NonSbuMoleculeCollection" = None,  # noqa: F821
    forbidden_indices: Optional[Iterable[int]] = None,
    create_single_metal_bus: bool = False,
    check_dimensionality: bool = True,
    break_organic_nodes_: bool = True,
) -> NodeCollection:
    """Find the nodes in a MOF.

    Args:
        mof (MOF): moffragmentor MOF instance
        unbound_solvent (NonSbuMoleculeCollection): collection of unbound solvent molecules
        forbidden_indices (Optional[Iterable[int]]): indices of sites that should not be considered
        create_single_metal_bus (bool): if True, single metal nodes will be created
        check_dimensionality (bool): if True, rod nodes will be broken into single metal nodes
        break_organic_nodes_ (bool): if True, rod nodes will be broken into single metal nodes
            if they are mostly organic

    Returns:
        NodeCollection: collection of nodes
    """
    if forbidden_indices is None:
        forbidden_indices = []

    node_result = find_node_clusters(
        mof, unbound_solvent.indices, forbidden_indices=forbidden_indices
    )
    if create_single_metal_bus:
        # Rewrite the node result
        node_result = create_single_metal_nodes(mof, node_result)
    if check_dimensionality:
        # If we have nodes with dimensionality >0, we change these nodes to only contain the metal
        node_result = break_rod_nodes(mof, node_result)
    if break_organic_nodes_:
        # ToDo: This, of course, would also break prophyrin ...
        node_result = break_organic_nodes(node_result, mof)

    node_collection = create_node_collection(mof, node_result)

    return node_result, node_collection


def metal_and_branching_coplanar(node_idx, all_branching_idx, mof, tol=0.1):
    branching_idx = list(node_idx & all_branching_idx)
    coords = mof.frac_coords[list(node_idx) + branching_idx]
    points = Points(coords)
    return points.are_coplanar(tol=tol)


def might_be_porphyrin(node_indices, branching_idx, mof):
    metal_in_node = _get_metal_sublist(node_indices, mof.metal_indices)
    node_indices = set(node_indices)
    branching_idx = set(branching_idx)
    bound_to_metal = sum([mof.get_neighbor_indices(i) for i in metal_in_node], [])
    branching_bound_to_metal = branching_idx & set(bound_to_metal)
    # ToDo: check and think if this can handle the general case
    # it should, at least if we only look at the metals
    if (len(metal_in_node) == 1) & (len(node_indices | branching_bound_to_metal) > 1):
        logger.debug(
            "metal_in_node",
            metal_in_node,
            node_indices,
        )
        num_neighbors = len(bound_to_metal)
        if (
            metal_and_branching_coplanar(node_indices, branching_bound_to_metal, mof)
            & (num_neighbors > 2)
            & (len(branching_bound_to_metal) < 5)
        ):
            logger.debug(
                "Metal in linker found, indices: {}".format(
                    node_indices,
                )
            )
            return True

    return False


def detect_porphyrin(node_collection, mof):
    not_node = []
    for i, node in enumerate(node_collection):
        if might_be_porphyrin(node._original_indices, node._original_graph_branching_indices, mof):
            logger.info("Found porphyrin in node {}".format(node._original_indices))
            not_node.append(i)
    return not_node



# -*- coding: utf-8 -*-
"""Methods for the fragmentation of MOFs"""
from collections import namedtuple

from loguru import logger

from moffragmentor.fragmentor._no_core_linker import generate_new_node_collection
from moffragmentor.fragmentor.linkerlocator import create_linker_collection
from moffragmentor.fragmentor.nodelocator import detect_porphyrin, find_nodes
from moffragmentor.fragmentor.solventlocator import (
    get_all_bound_solvent_molecules,
    get_floating_solvent_molecules,
)
from moffragmentor.net import build_net
from moffragmentor.sbu.linkercollection import LinkerCollection

__all__ = ["FragmentationResult"]

FragmentationResult = namedtuple(
    "FragmentationResult",
    [
        "nodes",
        "linkers",
        "bound_solvent",
        "unbound_solvent",
        "capping_molecules",
        "net_embedding",
        "has_1d_sbu",
    ],
)

_MAX_LOOPS = 2


def run_fragmentation(
    mof,
    check_dimensionality: bool = True,
    create_single_metal_bus: bool = False,
    break_organic_nodes_at_metal: bool = True,
) -> FragmentationResult:
    """Take a MOF and split it into building blocks."""
    logger.debug("Starting fragmentation with location of unbound solvent")
    unbound_solvent = get_floating_solvent_molecules(mof)
    need_rerun = True
    forbidden_indices = []
    counter = 0
    has_1d_sbu = None
    while need_rerun:
        try:
            not_node = []
            logger.debug(f"Fragmenting MOF for the {counter} time")
            node_result, node_collection = find_nodes(
                mof,
                unbound_solvent,
                forbidden_indices,
                create_single_metal_bus,
                check_dimensionality,
                break_organic_nodes_at_metal,
            )

            # Find bound solvent
            logger.debug("Locating bound solvent")
            bound_solvent = get_all_bound_solvent_molecules(mof, node_result.nodes)
            logger.debug(f"Found bound solvent {len(bound_solvent.indices)>0}")

            logger.debug("Locating linkers")
            linker_collection = create_linker_collection(
                mof, node_result, node_collection, unbound_solvent, bound_solvent
            )
            logger.debug(f"Found {len(linker_collection)} linkers")
            # if we have no linker we need to rerun the node detection
            if len(linker_collection) == 0:
                logger.debug("No linkers found, rerunning node detection")
                create_single_metal_bus = True
                need_rerun = True
            else:
                logger.debug("Checking for metal in linker")
                # ToDo: factor this out into its own function

                not_node = detect_porphyrin(node_collection, mof)

                for node in not_node:
                    forbidden_indices.extend(list(node_collection[node]._original_indices))

                if len(not_node) == 0:
                    need_rerun = False
                    break
                if len(not_node) == len(node_result.nodes):
                    logger.warning(
                        "We have metal in plane with the organic part. \
                        Which would indicate a prophyrin. \
                        However, there is no other metal cluster, so we will treat it as metal cluster."
                    )
                    need_rerun = False
                    break

        except Exception as e:
            logger.exception(f"Error while fragmenting: {e}")
        finally:
            counter += 1
            if counter > _MAX_LOOPS:
                need_rerun = False
                raise ValueError(f"Could not fragment after {counter} attempts.")

    logger.debug(
        "Check if we need to move the capping molecules from the linkercollection into their own collection"
    )
    is_linker = []
    is_capping = []
    for linker in linker_collection:
        if len(linker._original_graph_branching_indices) >= 2:
            is_linker.append(linker)
        else:
            if len(linker.molecule) > 1:
                is_capping.append(linker)
            else:
                logger.warning(
                    "Capping molecule with only one atom. Perhaps something funny happened in the fragmentation."
                )

    linker_collection = LinkerCollection(is_linker)
    if len(linker_collection) == 0 and len(is_capping) > 0:
        logger.warning(
            "No linkers found, but capping molecules. Assigning capping molecules to linkers."
        )
        linker_collection = LinkerCollection(is_capping)
        is_capping = []

    capping_molecules = LinkerCollection(is_capping)

    # Now handle the case of the the frameworks that have linkers without core (e.g. H-COO)
    # we detect those by having 0 entries in the linker collection and all capping molecules overlapping with
    # with the node collection
    # in this case, we will pass the capping molecules to the net constructor
    use_capping_in_net = False
    if len(linker_collection) == 0:
        logger.warning("No linkers found")
        if all(set(c._original_indices) & node_collection.indices for c in capping_molecules):
            logger.warning(
                "All capping molecules overlap with node collection. Will use them for net construction."
            )
            logger.debug("Constructing the embedding")
            use_capping_in_net = True

    try:
        if use_capping_in_net:
            logger.debug("Constructing the embedding")
            # However, I'd also need to split the node in this case
            new_node_collection = generate_new_node_collection(mof, node_result)

            # ToDo: need to add some code as follows
            # Need to create also "netnodes" before that
            # egde_candiates = defaultdict(list)
            # for i, netnode_i in enumerate(netnodes):
            #     for j, netnode_j in enumerate(netnodes):
            #         at_least_one_edge, images = has_edge(netnode_i, netnode_j, lattice)
            #         if at_least_one_edge:
            #             for coord, image_a, image_b in images:
            #                 if i == j and all(image_a == (0,0,0)) and all(image_b == (0,0,0)):
            #                     continue
            #                 metal_center = lattice.get_fractional_coords(netnode_i._coords) + image_a
            #                 linker_center = lattice.get_fractional_coords(netnode_j._coords) + image_b
            #                 edge = VoltageEdge(
            #                     linker_center - metal_center,
            #                     i,
            #                     j,
            #                     image_a,
            #                     image_b,
            #                 )
            #                 egde_candiates[
            #                     (round(coord[0], 2), round(coord[1], 2), round(coord[2], 2))
            #                 ].append((edge, np.abs(image_b).sum()))

            net_embedding = build_net(new_node_collection, capping_molecules, mof.lattice)
        else:
            logger.debug("Constructing the embedding")
            # Now, get the net
            net_embedding = build_net(node_collection, linker_collection, mof.lattice)
    except Exception as e:
        logger.exception(f"Error {e} in net construction")
        net_embedding = None
    fragmentation_results = FragmentationResult(
        node_collection,
        linker_collection,
        bound_solvent,
        unbound_solvent,
        capping_molecules,
        net_embedding,
        has_1d_sbu,
    )

    return fragmentation_results



# -*- coding: utf-8 -*-
"""Searches, such as depth-first search (DFS) on graphs."""
from copy import deepcopy
from typing import Dict, List, Tuple

import networkx as nx

from ..utils import _flatten_list_of_sets


def _has_path_to_any_other_metal(mof, index: int, this_metal_index: int) -> bool:
    """
    Check if some neighbor at index is only connected to this_metal_index.

    Or if we have some path to some other metal.

    A valid linker will have a path to another metal
    wheras a solvent molecule will only have a bond to `this_metal_index`

    Args:
        mof (MOF): A MOF instance that must provide
            .metal_indices and ._undirected_graph
        index (int): Index of the neighbor of interest
        this_metal_index (int): Index of the metal the neighbor is bound to

    Returns:
        bool: True if there is some path to some other metal
    """
    metal_indices = deepcopy(mof.metal_indices)
    metal_indices.remove(this_metal_index)
    graph = deepcopy(mof.nx_graph)
    graph.remove_node(this_metal_index)
    for metal_index in metal_indices:
        if nx.has_path(graph, index, metal_index):
            return True
    return False


def recursive_dfs_until_terminal(
    mof,
    start: int,
    path: List[int] = None,  # noqa: B006
    skip_list: List[int] = None,  # noqa: B006
) -> List[int]:
    """From a given starting point perform depth-first search until leaf nodes are reached.

    Args:
        mof (MOF): A MOF instance
        start (int): Starting index for the search
        path (List[int]): Starting path. Defaults to [].
        skip_list (List[int]): List of indices to skip. Defaults to [].

    Returns:
        List[int]: Path between start and the leaf node
    """
    if path is None:
        path = []
    if skip_list is None:
        skip_list = []
    if (start not in path) and (start not in skip_list):
        path.append(start)
        if mof._is_terminal(start):  # pylint:disable=protected-access
            return path

        for neighbour in mof.get_neighbor_indices(start):
            path = recursive_dfs_until_terminal(mof, neighbour, path, skip_list)

    return path


def _complete_graph(
    mof, paths: List[List[int]], branching_nodes: List[List[int]]
) -> Tuple[List[List[int]], Dict[int, List]]:
    """Loop over all paths that were traversed in DFS.

    Then, if the verices are not branching add all the paths
    that lead to terminal nodes. That is, in this step, we're adding
    the "capping" components of MOF nodes like capping formates or
    µ(3)-OH centers.

    Args:
        mof (MOF): Instance of a MOF object
        paths (List[List[int]]): Paths traversed in the initial DFS
            from metal indices to branching nodes
        branching_nodes (List[List[int]]): nodes that were
            identified as branching nodes

    Returns:
        Tuple[List[List[int]], Dict[int, List[int]]]: completed graphs, paths to terminal nodes
            from branching nodes
    """
    completed_edges = []
    visited = set()
    # ToDo: Why to we do this sum thousand times in this code?
    branching_nodes = sum(branching_nodes, [])

    all_directions_to_terminal = {}
    for path in paths:
        branching_node_in_path = set(path) & set(
            branching_nodes
        )  # we need to expose those directions to complete, as we should also show them on linker/capping molecule
        directions_to_complete = []

        for branching_node in branching_node_in_path:
            for neighbour in mof.get_neighbor_indices(branching_node):
                if mof._leads_to_terminal((branching_node, neighbour)):
                    directions_to_complete.extend(
                        recursive_dfs_until_terminal(mof, neighbour, [], [branching_node])
                        + [branching_node, neighbour]
                    )
                if mof._is_terminal(neighbour):
                    directions_to_complete.extend([branching_node, neighbour])
        all_directions_to_terminal[branching_node] = directions_to_complete
        subpath = []
        for vertex in path:
            # ToDo: we can at least leverage the information about the bridges
            # we have and not just brute-force the search for all vertices
            if vertex not in visited:
                # ToDo: exclude those branching nodes that have one neighbor that leads to a terminal node
                p = recursive_dfs_until_terminal(mof, vertex, [], branching_nodes)
                subpath.extend(p)
        completed_edges.append(subpath + path + directions_to_complete)
    return completed_edges, all_directions_to_terminal


def recursive_dfs_until_branch(
    mof,
    start: int,
    path: List[int] = None,  # noqa: B006
    branching_nodes: List[int] = None,  # noqa: B006
) -> List[int]:
    """From a starting point perform DFS until branch nodes are reached.

    Args:
        mof (MOF): A MOF instance
        start (int): Starting index for the search
        path (List[int], optional): Starting path. Defaults to [].
        branching_nodes (List[int], optional): List of branching nodes.
            Defaults to [].

    Returns:
        List[int]: Path between start and the leaf node
    """
    if path is None:
        path = []
    if branching_nodes is None:
        branching_nodes = []
    if start not in path:
        path.append(start)

        if mof._is_branch_point(start):  # pylint:disable=protected-access
            branching_nodes.append(start)
            return path, branching_nodes

        for neighbour in mof.get_neighbor_indices(start):
            path, branching_nodes = recursive_dfs_until_branch(
                mof, neighbour, path, branching_nodes
            )

    return path, branching_nodes


def recursive_dfs_until_cn3(
    mof,
    start: int,
    path: List[int] = None,  # noqa: B006
    branching_nodes: List[int] = None,  # noqa: B006
) -> List[int]:
    """From starting point perform DFS until CN=3 nodes are reached.

    Args:
        mof (MOF): A MOF instance
        start (int): Starting index for the search
        path (List[int], optional): Starting path. Defaults to [].
        branching_nodes (List[int], optional): List of branching nodes.
            Defaults to [].

    Returns:
        List[int]: Path between start and the leaf node
    """
    if path is None:
        path = []
    if branching_nodes is None:
        branching_nodes = []
    if start not in path:
        path.append(start)

        if (len(mof.get_neighbor_indices(start)) >= 3) & (
            start not in mof.metal_indices
        ):  # pylint:disable=protected-access
            branching_nodes.append(start)
            return path, branching_nodes

        for neighbour in mof.get_neighbor_indices(start):
            path, branching_nodes = recursive_dfs_until_cn3(mof, neighbour, path, branching_nodes)

    return path, branching_nodes


def _to_graph(mof, paths, branch_sites):
    """https://stackoverflow.com/questions/4842613/merge-lists-that-share-common-elements"""
    graph = nx.Graph()
    for part in paths:
        graph.add_nodes_from(part)
        graph.add_edges_from(_to_edges(part))
    graph.add_edges_from(
        _connect_connected_branching_indices(mof, _flatten_list_of_sets(branch_sites))
    )
    return graph


def _to_edges(paths):
    iterator = iter(paths)
    last = next(iterator)  # pylint:disable=stop-iteration-return

    for current in iterator:
        yield last, current
        last = current


def _connect_connected_branching_indices(mof, flattend_path) -> set:
    edges = set()

    for i in flattend_path:
        neighbors = mof.get_neighbor_indices(i)
        for neighbor in neighbors:
            if neighbor in flattend_path:
                edges.add(tuple(sorted((i, neighbor))))

    return edges



# -*- coding: utf-8 -*-
"""Describing the organic building blocks, i.e., linkers."""
from copy import deepcopy

from pymatgen.core import Molecule
from rdkit.Chem import AllChem

from .sbu import SBU


class Linker(SBU):
    """Describe a linker in a MOF"""

    def __repr__(self) -> str:
        """Return a string representation of the linker."""
        return f"Linker ({self.composition})"


def _get_edge_dict_from_rdkit_mol(mol):
    edges = {}
    for bond in mol.GetBonds():
        edges[(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())] = None
    return edges


def _make_mol_from_rdkit_mol(mol):
    """Take the first conformer"""
    molecule = deepcopy(mol)
    AllChem.EmbedMolecule(molecule)  # pylint:disable=no-member
    # not clear to me why pylint complains
    conf = molecule.GetConformers()
    positions = conf[0].GetPositions()
    symbols = [atom.GetSymbol() for atom in mol.GetAtoms()]
    return Molecule(symbols, positions)



# -*- coding: utf-8 -*-
"""Make sure that we have no regression
in the output of the fragmentation"""
import os

import pytest

from moffragmentor import MOF

THIS_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_FILE_DIR = os.path.join(THIS_DIR, "..", "test_files")


def fragment(cif):
    path = os.path.join(TEST_FILE_DIR, cif)
    mof = MOF.from_cif(path)
    parts = mof.fragment()
    return parts


@pytest.mark.slow
def test_fragmentation():
    """On a set of structures from the review and beyond."""
    vejhez_parts = fragment("VEJHEZ_clean.cif")
    assert vejhez_parts.nodes.smiles[0] == "[C]1O[Zn]2O[C]O[Zn](O1)O[C]O2"
    assert len(set(vejhez_parts.nodes.smiles)) == 1

    assert (
        vejhez_parts.linkers.smiles[0]
        == "[O]C(=O)c1cc(cc(c1)C1=C2C=CC3=[N]2[Zn@]24n5c1ccc5C(=C1[N]2=C(C=C1)C(=c1n4c(=C3c2cc(cc(c2)C(=O)[O])C(=O)[O])cc1)c1cc(cc(c1)C(=O)[O])C(=O)[O])c1cc(cc(c1)C(=O)[O])C(=O)[O])C(=O)[O]"  # noqa: E501
    )
    assert len(set(vejhez_parts.linkers.smiles)) == 1
    assert vejhez_parts.net_embedding.rcsr_code == "the"

    hkust_parts = fragment("HKUST-1.cif")
    assert hkust_parts.linkers[0].smiles == "[O]C(=O)c1cc(C([O])=O)cc(C([O])=O)c1"

    assert hkust_parts.nodes[0].smiles == "[C]1O[Cu]234O[C]O[Cu]2(O1)(O[C]O3)O[C]O4"

    two_dimensional_mof_parts = fragment("2dmof.cif")
    assert len(two_dimensional_mof_parts.linkers.unique_sbus) == 1
    assert len(two_dimensional_mof_parts.nodes.unique_sbus) == 1



# -*- coding: utf-8 -*-
"""Helper functions."""
import datetime
import json
import os
import pickle
import sys
from collections import defaultdict
from shutil import which
from typing import Collection, Dict, Iterable, List, Union

import networkx as nx
import numpy as np
import pymatgen
from backports.cached_property import cached_property
from loguru import logger
from pymatgen.analysis.graphs import MoleculeGraph, StructureGraph
from pymatgen.core import Molecule, Structure
from pymatgen.io.babel import BabelMolAdaptor
from skspatial.objects import Points


def enable_logging() -> List[int]:
    """Set up the mofdscribe logging with sane defaults."""
    logger.enable("mofdscribe")

    config = dict(
        handlers=[
            dict(
                sink=sys.stderr,
                format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS Z UTC}</>"
                " <red>|</> <lvl>{level}</> <red>|</> <cyan>{name}:{function}:{line}</>"
                " <red>|</> <lvl>{message}</>",
                level="INFO",
            ),
            dict(
                sink=sys.stderr,
                format="<red>{time:YYYY-MM-DD HH:mm:ss.SSS Z UTC} | {level} | {name}:{function}:{line} | {message}</>",
                level="WARNING",
            ),
        ]
    )
    return logger.configure(**config)


def get_molecule_mass(molecule):
    mass = 0
    for site in molecule:
        mass += site.specie.atomic_mass
    return mass


def unwrap(positions, lattice):
    celldiag = np.diagonal(lattice.matrix)
    dxyz = positions - positions[0]
    dxyz -= celldiag * np.around(dxyz / celldiag)
    dxyz += positions[0]

    return dxyz


def _get_metal_sublist(
    indices: List[int],
    metal_indices: List[int],
    periodic_index_map: Union[dict, None] = None,
) -> List[int]:
    metal_subset_in_node = []

    for node_index in indices:
        if node_index in metal_indices:
            if periodic_index_map is not None:
                # adding the periodic images
                metal_subset_in_node.extend(periodic_index_map[node_index])
            else:
                metal_subset_in_node.append(node_index)

    return metal_subset_in_node


def _get_metal_sublists(
    indices: List[List[int]],
    metal_indices: list,
    periodic_index_map: Union[dict, None] = None,
) -> List[List[int]]:
    """Recover the metal fragments from the nodes.

     We need it, for example, in the node filtering step where we analyze
     if the removal of a node creates new connected components.

    Args:
        indices (List[List[int]]): input indices, e.g., node indices
        metal_indices (list): indices of the metals in the structure
        periodic_index_map (dict): If not None, then we will also
            add the periodic images according to this map.
            Defaults to None.

    Returns:
        List[List[int]]: filtered input list,
            that now only contains indices of atoms that are metals
    """
    output_list = []
    for sublist in indices:
        new_sublist = _get_metal_sublist(sublist, metal_indices, periodic_index_map)
        output_list.append(new_sublist)
    return output_list


def revert_dict(dictionary):
    new_d = {}

    for key, value in dictionary.items():
        new_d[value] = key

    return new_d


def get_smiles_from_pmg_mol(pmg_mol):
    from openbabel import pybel as pb  # pylint:disable=import-outside-toplevel

    adaptor = BabelMolAdaptor(pmg_mol)
    pm = pb.Molecule(adaptor.openbabel_mol)
    smiles = pm.write("can").strip()
    return smiles


def mol_from_sites(sites):
    coords = [n.coords for n in sites]
    species = [n.specie for n in sites]
    binding = [n.properties["binding"] for n in sites]

    molecule = Molecule(species, coords, site_properties={"binding": binding})
    return molecule.get_centered_molecule()


def build_molecule_and_graph(s, sites, edges):
    selected_sites = [s[i] for i in sites]

    remapper = dict(zip(sites, np.arange(len(sites))))
    new_edges = {}

    for a, b in edges:
        new_edges[(remapper[a], remapper[b])] = None

    mol = mol_from_sites(selected_sites)
    molecule_graph = MoleculeGraph.with_edges(mol, new_edges)
    return mol, molecule_graph


def write_cif(  # pylint:disable=too-many-locals
    s, graph, connection_indices, molecule=None, write_bonding_mode=False
):
    header_lines = [
        "data_cif",
        "_audit_creation_date              " + datetime.datetime.today().strftime("%Y-%m-%d"),
        "_audit_creation_method            'moffragmentor'",
        "_symmetry_space_group_name_H-M    'P 1'",
        "_symmetry_Int_Tables_number       1",
        "_symmetry_cell_setting            triclinic",
        "loop_",
        "_symmetry_equiv_pos_as_xyz",
        "  x,y,z",
    ]

    cell_lines = [
        "_cell_length_a                    " + str(s.lattice.a),
        "_cell_length_b                    " + str(s.lattice.b),
        "_cell_length_c                    " + str(s.lattice.c),
        "_cell_angle_alpha                 " + str(s.lattice.alpha),
        "_cell_angle_beta                  " + str(s.lattice.beta),
        "_cell_angle_gamma                 " + str(s.lattice.gamma),
    ]

    loop_header = [
        "loop_",
        "_atom_site_label",
        "_atom_site_type_symbol",
        "_atom_site_fract_x",
        "_atom_site_fract_y",
        "_atom_site_fract_z",
        "_atom_site_charge",
    ]

    loop_content = []
    site_index = {}
    for i, site in enumerate(s):
        vec = site.frac_coords
        es = str(site.specie)

        if i in connection_indices:
            ind = "X" + str(i)
            loop_content.append(f" {ind} {es}  {vec[0]:.6f}  {vec[1]:.6f}  {vec[2]:.6f}  {0:.6f}")
        else:
            ind = es + str(i)
            loop_content.append(f" {ind} {es}  {vec[0]:.6f}  {vec[1]:.6f}  {vec[2]:.6f}  {0:.6f}")
        site_index[i] = ind

    connection_loop_header = [
        "loop_",
        "_geom_bond_atom_site_label_1",
        "_geom_bond_atom_site_label_2",
    ]

    if write_bonding_mode:

        connection_loop_header += ["_geom_bond_distance", "_ccdc_geom_bond_type"]

    connection_loop_content = []
    set_bond = set()
    if molecule:
        s = molecule
    for i, _ in enumerate(s):
        neighbors = graph.get_connected_sites(i)
        for j, neighbor_site in enumerate(neighbors):
            ind0 = site_index[i]
            ind1 = site_index[neighbor_site.index]

            tuple_a = (ind0, ind1)
            tuple_b = (ind1, ind0)

            if not (tuple_a in set_bond) and not (
                tuple_b in set_bond
            ):  # pylint:disable=superfluous-parens
                set_bond.add(tuple_a)

                dist = np.round(s.get_distance(i, j), 3)
                if write_bonding_mode:
                    connection_loop_content.append(f"{ind0:7} {ind1:>7} {dist:>7} {'.':>3} {s:>3}")
                else:
                    connection_loop_content.append(f"{ind0:7} {ind1:>7} {'':>7} {'':>3}")

    return "\n".join(
        header_lines
        + cell_lines
        + loop_header
        + loop_content
        + connection_loop_header
        + connection_loop_content
    )


def pickle_dump(obj, path):
    with open(path, "wb") as handle:
        pickle.dump(obj, handle)


def make_if_not_exists(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def is_tool(name: str) -> bool:
    """Check whether `name` is on PATH and marked as executable.

    https://stackoverflow.com/questions/11210104/check-if-a-program-exists-from-a-python-script

    Args:
        name (str): The name of the tool to check for.

    Returns:
        bool: True if the tool is on PATH and marked as executable.
    """
    return which(name) is not None


def _not_relevant_structure_indices(structure: Structure, indices: Collection[int]) -> List[int]:
    """Return the indices of the structure that are *not* in the indices collection.

    Args:
        structure (Structure): pymatgen Structure object
        indices (Collection[int]): Collection of integers

    Returns:
        List[int]: indices of structure that are not in the indices collection
    """
    not_relevant = []
    for i in range(len(structure)):
        if i not in indices:
            not_relevant.append(i)
    return not_relevant


def get_sub_structure(mof: "MOF", indices: Collection[int]) -> Structure:  # noqa: F821
    """Return a sub-structure of the structure with only the sites with the given indices.

    Args:
        mof (MOF): MOF object
        indices (Collection[int]): Collection of integers

    Returns:
        Structure: sub-structure of the structure with only the sites with the given indices
    """
    sites = []
    for index in indices:
        sites.append(mof.structure[index])

    s = Structure.from_sites(sites, to_unit_cell=True)
    return s


def visualize_part(mof, indices: Collection):
    import nglview  # pylint:disable=import-outside-toplevel

    sites = []
    for index in indices:
        sites.append(mof.structure[index])

    s = Structure.from_sites(sites, to_unit_cell=True)

    return nglview.show_pymatgen(s)


def joint_visualize_mols(mols: Iterable[Molecule]):
    import nglview as nv

    sites = []
    for mol in mols:
        sites.extend(mol.sites)
    super_mol = Molecule.from_sites(sites)
    return nv.show_pymatgen(super_mol)


def _flatten_list_of_sets(parts):
    flattened_parts = []

    for sublist in parts:
        for elm in sublist:
            flattened_parts.append(elm)
    return flattened_parts


def _get_image(mof, reference_index: int, index: int):
    _, image = mof.structure.lattice.get_distance_and_image(
        mof.frac_coords[reference_index], mof.frac_coords[index]
    )
    return image


def _get_cartesian_coords(mof, reference_index: int, index: int):
    image = _get_image(mof, reference_index, index)
    position = mof.frac_coords[index] + image
    return mof.lattice.get_cartesian_coords(position)


def _get_molecule_edge_label(u: int, v: int):
    return tuple(sorted((u, v)))


def _reindex_list_of_tuple(list_of_tuples, remapping: Dict[int, int]):
    new_list = []
    for tupl in list_of_tuples:
        new_tuple = (remapping[tupl[0]], remapping[tupl[1]])
        new_list.append(new_tuple)
    return new_list


def _metal_in_edge(site_collection, edge):
    return any(site_collection[edge_partner].specie.is_metal for edge_partner in edge)


def _get_vertices_of_smaller_component_upon_edge_break(graph, edge):
    graph_copy = graph.copy()
    graph_copy.remove_edge(edge[0], edge[1])
    connected_components = nx.connected_components(graph_copy)
    smallest_connected_component = min(connected_components, key=len)
    return smallest_connected_component


def get_edge_dict(structure_graph: StructureGraph) -> dict:
    def get_label(u, v):
        return sorted((u, v))

    types = defaultdict(list)
    for u, v, _ in structure_graph.graph.edges(data=True):
        label = get_label(u, v)
        types[tuple(label)] = None
    return dict(types)


def get_linker_connectivity(edge_dict):
    linker_connectivity = defaultdict(set)
    for node, linkers in edge_dict.items():
        for linker_tuple in linkers:
            linker_connectivity[linker_tuple[0]].add(node)
    num_neighbors = {}
    for linker, neighbors in linker_connectivity.items():
        num_neighbors[linker] = len(neighbors)

    return num_neighbors


def get_neighbors_from_nx_graph(graph, node_idx):
    return list(nx.neighbors(graph, node_idx))


def remove_all_nodes_not_in_indices(graph: nx.Graph, indices) -> nx.Graph:
    to_delete = [i for i in range(len(graph)) if i not in indices]
    graph.structure = Structure.from_sites(graph.structure.sites)
    graph.remove_nodes(to_delete)


class IStructure(pymatgen.core.structure.IStructure):
    """pymatgen IStructure with faster equality comparison.

    This dramatically speeds up lookups in the LRU cache when an object
    with the same __hash__ is already in the cache.
    """

    __hash__ = pymatgen.core.structure.IStructure.__hash__

    def __eq__(self, other):
        """Use specific, yet performant hash for equality comparison."""
        return self._dict_hash == other._dict_hash

    @cached_property
    def _dict_hash(self):
        """Specific, yet performant hash."""
        return hash(json.dumps(self.as_dict(), sort_keys=True))


def remove_site(structure: Union[Structure, IStructure]) -> None:
    if isinstance(structure, IStructure):
        structure = Structure.from_sites(structure.sites)


def add_suffix_to_dict_keys(dictionary: dict, suffix: str) -> dict:
    return {f"{k}_{suffix}": v for k, v in dictionary.items()}


def are_coplanar(mof, indices, tol=0.1):
    coords = mof.frac_coords[indices]
    points = Points(coords)
    return points.are_coplanar(tol=tol)



# -*- coding: utf-8 -*-
"""Collection for MOF building blocks"""
from collections import Counter
from typing import List

from backports.cached_property import cached_property

from .sbu import SBU


class SBUCollection:
    """Container for a collection of SBUs"""

    def __init__(self, sbus: List[SBU]):
        """Construct a SBUCollection.

        Args:
            sbus (List[SBU]): List of SBU objects.

        Raises:
            ValueError: If there is an unexpected number of SBUs
                (inconsistency in the extracted indices and the SBUs).
        """
        self.sbus = sbus
        self._sbu_types = None
        self._composition = None
        self._unique_sbus = None
        self._centers = [sbu.center for sbu in self.sbus]
        self._indices = [sbu.get_indices() for sbu in self.sbus]

        if len(self._indices) != len(self.sbus):
            raise ValueError("Number of SBUs is inconsistent")

    def __len__(self) -> int:
        """Return number of SBUs."""
        return len(self.sbus)

    def __getitem__(self, index) -> SBU:
        """Return SBU at index."""
        return self.sbus[index]

    def __iter__(self):
        """Iterate over SBUs (generator)."""
        for sbu in self.sbus:
            yield sbu

    @property
    def indices(self):
        return set(sum(self._indices, []))

    @property
    def centers(self):
        return self._centers

    @cached_property
    def smiles(self):
        """
        Return a list of the SMILES strings of the SBUs.

        Returns:
            List[str]: A list of smiles strings.
        """
        return [sbu.smiles for sbu in self.sbus]

    @property
    def sbu_types(self):
        if not self._sbu_types:
            self._get_unique()
        return self._sbu_types

    @property
    def coordination_numbers(self):
        return [sbu.coordination for sbu in self.sbus]

    @property
    def molar_masses(self):
        return [sbu.molar_mass for sbu in self.sbus]

    @property
    def unique_sbus(self):
        if not self._sbu_types:
            self._get_unique()
        return self._unique_sbus

    @property
    def sbu_properties(self):
        return dict(zip(self.sbu_types, self.coordination_numbers))

    def _get_unique(self):
        all_strings = [str(sbu) for sbu in self.sbus]
        unique_strings = set(all_strings)
        unique_mapping = []
        for string in all_strings:
            for i, unique_string in enumerate(unique_strings):
                if string == unique_string:
                    unique_mapping.append(i)
                    break
        self._unique_sbus = unique_strings
        self._sbu_types = unique_mapping

    def _get_composition(self):
        if self._composition is None:
            composition = []
            for mol in self.sbus:
                composition.append(str(mol.composition))
            composition_counter = Counter(composition)
            self._composition = dict(composition_counter)

        return self._composition

    @property
    def composition(self):
        return self._get_composition()

    def __repr__(self) -> str:
        """Return string representation."""
        return f"SBUCollection({self.composition})"



# -*- coding: utf-8 -*-
"""Funnctions that can be used to locate bound and unbound sovlent"""
from collections import OrderedDict
from typing import List, Set

from structuregraph_helpers.subgraph import get_subgraphs_as_molecules

from ._graphsearch import _has_path_to_any_other_metal, recursive_dfs_until_terminal
from .filter import bridges_across_cell
from ..molecule import NonSbuMolecule, NonSbuMoleculeCollection
from ..utils import _get_metal_sublist

__all__ = ["get_floating_solvent_molecules", "get_all_bound_solvent_molecules"]


def get_floating_solvent_molecules(mof) -> NonSbuMoleculeCollection:
    """Create a collection of NonSbuMolecules from a MOF.

    Args:
        mof (MOF): instance of MOF

    Returns:
        NonSbuMoleculeCollection: collection of NonSbuMolecules
    """
    mols, graphs, idx, _, _ = get_subgraphs_as_molecules(mof.structure_graph, return_unique=False)
    molecules = []

    for mol, graph, identifier in zip(mols, graphs, idx):
        molecules.append(NonSbuMolecule(mol, graph, identifier))

    return NonSbuMoleculeCollection(molecules)


def _get_solvent_molecules_bound_to_node(mof, node_atoms: Set[int]) -> NonSbuMoleculeCollection:
    """Locate solvent molecules bound to one MOF node.

    Bound solvent is defined as being connected via one bridge
    to one metal center.

    Args:
        mof (MOF): Instance of a MOF object
        node_atoms (Set[int]): Indices that were identified as
            belonging to a node cluster.

    Returns:
        NonSbuMoleculeCollection: Collection of NonSbuMolecule objects
            containing the bound solvent molecules
    """
    bound_solvent_location_result = _locate_bound_solvent(mof, node_atoms)

    molecules = []

    for solvent_ind in bound_solvent_location_result["solvent_indices"]:
        # check if the solvent candidate has bridge to two different metals:
        # if so, it is not a solvent
        # that is, we consider capping groups different from bound solvent
        neighbors: Set[int] = set()
        for ind in solvent_ind:
            # this is here actually not enough.
            # we also need to look at the images
            neighbors.update(mof.get_neighbor_indices(ind))

        if not len(neighbors.intersection(mof.metal_indices)) > 1:
            molecules.append(
                NonSbuMolecule.from_structure_graph_and_indices(mof.structure_graph, solvent_ind)
            )

    return NonSbuMoleculeCollection(molecules)


def get_all_bound_solvent_molecules(
    mof, node_atom_sets: List[Set[int]]
) -> NonSbuMoleculeCollection:
    """Identify all bound solvent molecules.

    Bound solvent is defined as being connected via one bridge
    to one metal center.

    Args:
        mof (MOF): instance of a MOF object
        node_atom_sets (List[Set[int]]): List of indices for the MOF nodes

    Returns:
        NonSbuMoleculeCollection: Collection of NonSbuMolecule objects
            containing the bound solvent molecules
    """
    non_sbu_molecule_collections = NonSbuMoleculeCollection([])
    for node_atom_set in node_atom_sets:
        non_sbu_molecule_collections += _get_solvent_molecules_bound_to_node(mof, node_atom_set)

    return non_sbu_molecule_collections


def find_solvent_molecule_indices(mof, index: int, starting_metal: int) -> List[int]:
    """Find all the indices that belong to a solvent molecule.

    Args:
        mof (MOF): index
        index (int): Starting index of solvent molecules
        starting_metal (int): Metal index to which solvent is bound to

    Returns:
        List[int]: List of indices that belong to solvent molecule
    """
    path = recursive_dfs_until_terminal(mof, index, [starting_metal])
    _ = path.pop(0)
    return path


def _locate_bound_solvent(mof, node_atoms: Set[int]) -> OrderedDict:
    solvent_connections = set()

    solvent_indices = []
    good_connections = set()

    metal_subset = _get_metal_sublist(node_atoms, mof.metal_indices)

    for metal_index in metal_subset:
        metal_neighbors = mof.get_neighbor_indices(metal_index)

        for metal_neighbor in metal_neighbors:
            # Note: we cannot do a simple bridge check here because we would not find things such as acetate
            if not _has_path_to_any_other_metal(mof, metal_neighbor, metal_index):
                potential_solvent_indices = find_solvent_molecule_indices(
                    mof, metal_neighbor, metal_index
                )
                if not bridges_across_cell(mof, potential_solvent_indices):
                    solvent_connections.add(metal_neighbor)
                    solvent_indices.append(potential_solvent_indices)

            else:
                good_connections.add(metal_neighbor)

    return OrderedDict(
        [
            ("solvent_indices", solvent_indices),
            ("solvent_connections", solvent_connections),
            ("non_solvent_connections", good_connections),
        ]
    )



# -*- coding: utf-8 -*-
"""Create Python containers for node building blocks.

Here we understand metal clusters as nodes.
"""
from collections import defaultdict
from typing import Set

from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.core import Molecule, Site, Structure

from moffragmentor.fragmentor.molfromgraph import wrap_molecule
from moffragmentor.fragmentor.splitter import _sites_and_classified_indices_from_indices
from moffragmentor.sbu.sbu import SBU
from moffragmentor.utils import _reindex_list_of_tuple

__all__ = ["Node"]

_BINDING_DUMMY = "Xe"
_BRANCHING_DUMMY = "Kr"


def _build_mol_and_graph(mof, indices, ignore_hidden_indices=True, add_additional_site=True):
    indices = set(indices)
    sites_and_indices = _sites_and_classified_indices_from_indices(mof, indices)
    if ignore_hidden_indices:
        relevant_indices = indices - set(sites_and_indices.hidden_vertices)
    else:
        relevant_indices = indices

    selected_positions, selected_species, selected_edge_indices = [], [], []

    # give us a quick and easy way to filter the edges we want to keep
    # for this reason we will use both edge "partners" as keys
    # and have a list of indices in the original edges list as the value
    edge_dict = defaultdict(list)
    for i, egde in enumerate(sites_and_indices.edges):
        edge_partner_1, edge_partner_2 = egde
        edge_dict[edge_partner_1].append(i)
        edge_dict[edge_partner_2].append(i)

    for i, index in enumerate(indices):
        if index in relevant_indices:
            selected_positions.append(sites_and_indices.cartesian_coordinates[i])
            selected_species.append(sites_and_indices.species[i])
            edge_idxs = edge_dict[index]
            for edge_idx in edge_idxs:
                edge = sites_and_indices.edges[edge_idx]
                if (edge[0] in relevant_indices) and (edge[1] in relevant_indices):
                    selected_edge_indices.append(edge_idx)

    index_map = dict(zip(relevant_indices, range(len(relevant_indices))))

    molecule = Molecule(selected_species, selected_positions)
    selected_edges = [sites_and_indices.edges[i] for i in selected_edge_indices]
    selected_edges = _reindex_list_of_tuple(selected_edges, index_map)
    edge_dict = dict(zip(selected_edges, [None] * len(selected_edges)))
    molecule_graph = MoleculeGraph.with_edges(molecule, edge_dict)

    mol, mapping = wrap_molecule(
        list(relevant_indices), mof, add_additional_site=add_additional_site
    )

    return mol, molecule_graph, mapping


def _extract_and_wrap(node_indices, all_branching_indices, all_binding_indices, mof):
    # first, create a molecule and molecule graph from those indices
    # that intersect with the node indices
    branching_in_node = [i for i in all_branching_indices if i in node_indices]
    binding_in_node = [i for i in all_binding_indices if i in node_indices]

    mol, graph, mapping = _build_mol_and_graph(
        mof, list(node_indices) + branching_in_node + binding_in_node
    )

    # now, also create one molecule from the node indices, where we replace
    # binding and branching indices with dummy atoms
    node_metal_atoms = [i for i in node_indices if i in mof.metal_indices]
    binding_to_node_metal = set(sum([mof.get_neighbor_indices(i) for i in node_metal_atoms], []))
    binding_to_node_metal = binding_to_node_metal.intersection(all_binding_indices)

    branching_to_node_metal = set(
        sum(
            [mof.get_neighbor_indices(i) for i in node_metal_atoms if i in all_branching_indices],
            [],
        )
    )
    binding_neighbors = sum([mof.get_neighbor_indices(i) for i in binding_to_node_metal], [])
    branching_to_binding = set(binding_neighbors).intersection(all_branching_indices)

    # Now, make a copy of the structure and replace the indices with dummy atoms
    dummy_structure = Structure.from_sites(mof.structure.sites)
    for i in binding_to_node_metal:
        dummy_structure.replace(i, _BINDING_DUMMY)
    for i in branching_to_binding | branching_to_node_metal:
        dummy_structure.replace(i, _BRANCHING_DUMMY)
    dummy_branching_sites = branching_to_binding | branching_to_node_metal
    mol_w_dummy, graph_w_dummy, mapping_w_dummy = _build_mol_and_graph(
        mof,
        list(node_indices)
        + list(binding_to_node_metal)
        + list(branching_to_binding)
        + list(branching_to_node_metal),
        ignore_hidden_indices=False,
        add_additional_site=False,
    )

    inverse_mapping = {v[0]: k for k, v in mapping_w_dummy.items()}
    # let's replace here now the atoms with the dummys as doing it beforehand might cause issues
    # (e.g. we do not have the distances for a cutoffdict)

    for i in branching_to_binding | branching_to_node_metal:
        if i not in node_indices & set(mof.metal_indices):
            mol_w_dummy._sites[inverse_mapping[i]] = Site(
                _BRANCHING_DUMMY,
                mol_w_dummy._sites[inverse_mapping[i]].coords,
                properties={"original_species": str(mol_w_dummy._sites[inverse_mapping[i]].specie)},
            )

    for i in binding_to_node_metal:
        mol_w_dummy._sites[inverse_mapping[i]] = Site(
            _BINDING_DUMMY,
            mol_w_dummy._sites[inverse_mapping[i]].coords,
            properties={"original_species": str(mol_w_dummy._sites[inverse_mapping[i]].specie)},
        )

    graph_w_dummy.molecule = mol_w_dummy

    return mol, graph, mapping, mol_w_dummy, graph_w_dummy, mapping_w_dummy, dummy_branching_sites


def node_from_mof_and_indices(cls, mof, node_indices, all_branching_indices, all_binding_indices):
    """Create a node from a MOF and a list of indices of different types."""
    node_indices = node_indices
    branching_indices = node_indices & all_branching_indices
    binding_indices = node_indices & all_binding_indices
    graph_branching_indices = branching_indices & node_indices

    (
        mol,
        graph,
        mapping,
        mol_w_dummy,
        graph_w_dummy,
        mapping_w_dummy,
        dummy_branching_sites,
    ) = _extract_and_wrap(
        node_indices=node_indices,
        all_branching_indices=all_branching_indices,
        all_binding_indices=all_binding_indices,
        mof=mof,
    )

    node = cls(
        molecule=mol,
        molecule_graph=graph,
        graph_branching_indices=graph_branching_indices,
        binding_indices=binding_indices,
        molecule_original_indices_mapping=mapping,
        dummy_molecule=mol_w_dummy,
        dummy_molecule_graph=graph_w_dummy,
        dummy_molecule_indices_mapping=mapping_w_dummy,
        dummy_branching_indices=dummy_branching_sites,
    )

    return node


class Node(SBU):
    """Container for metal cluster building blocks.

    Will typically automatically be constructured by the fragmentor.
    """

    @classmethod
    def from_mof_and_indices(  # pylint:disable=too-many-arguments
        cls,
        mof,
        node_indices: Set[int],
        branching_indices: Set[int],
        binding_indices: Set[int],
    ):
        """Build a node object from a MOF and some intermediate outputs of the fragmentation.

        Args:
            mof (MOF): The MOF to build the node from.
            node_indices (Set[int]): The indices of the nodes in the MOF.
            branching_indices (Set[int]): The indices of the branching points in the MOF
                that belong to this node.
            binding_indices (Set[int]): The indices of the binding points in the MOF
                that belong to this node.

        Returns:
            A node object.
        """
        return node_from_mof_and_indices(
            cls,
            mof,
            node_indices,
            all_branching_indices=branching_indices,
            all_binding_indices=binding_indices,
        )

    def __repr__(self) -> str:
        """Return a string representation of the node."""
        return f"Node ({self.composition})"



# -*- coding: utf-8 -*-
"""Representation for a secondary building block."""
import warnings
from collections import defaultdict
from typing import Collection, Dict, List, Optional, Tuple

import networkx as nx
import numpy as np
import pubchempy as pcp
from backports.cached_property import cached_property
from loguru import logger
from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.core import Molecule, Structure
from pymatgen.io.babel import BabelMolAdaptor
from rdkit import Chem
from scipy.spatial.distance import pdist

from moffragmentor.utils import pickle_dump
from moffragmentor.utils.mol_compare import mcs_rank


def ob_mol_without_metals(obmol):
    """Remove metals from an OpenBabel molecule."""
    import openbabel as ob

    mol = obmol.clone
    for atom in ob.OBMolAtomIter(mol.OBMol):
        if atom.IsMetal():
            mol.OBMol.DeleteAtom(atom)

    return mol


__all__ = ["SBU"]


def obmol_to_rdkit_mol(obmol):
    """Convert an OpenBabel molecule to a RDKit molecule."""
    smiles = obmol.write("can").strip()
    mol = Chem.MolFromSmiles(smiles, sanitize=True)
    if mol is None:
        warnings.warn("Attempting to remove metals to generate RDKit molecule")
        new_obmol = ob_mol_without_metals(obmol)
        smiles = new_obmol.write("can").strip()
        mol = Chem.MolFromSmiles(smiles, sanitize=True)

    return mol


class SBU:
    """Representation for a secondary building block.

    It also acts as container for site indices:

        * graph_branching_indices: are the branching indices according
            to the graph-based definition. They might not be part of the molecule.
        * binding_indices: are the indices of the sites between
            the branching index and metal
        * original_indices: complete original set of indices that has been selected
            for this building blocks

    .. note::

        The coordinates in the molecule object are not the ones directly
        extracted from the MOF. They are the coordinates of sites unwrapped
        to ensure that there are no "broken molecules" .

        To obtain the "original" coordinates, use the `_coordinates` attribute.

    .. note:: Dummy molecules

        In dummy molecules the binding and branching sites are replaces by
        dummy atoms (noble gas). They also have special properties that indicate
        the original species.

    Examples:
        >>> # visualize the molecule
        >>> sbu_object.show_molecule()
        >>> # search pubchem for the molecule
        >>> sbu_object.search_pubchem()
    """

    def __init__(
        self,
        molecule: Molecule,
        molecule_graph: MoleculeGraph,
        graph_branching_indices: Collection[int],
        binding_indices: Collection[int],
        molecule_original_indices_mapping: Optional[Dict[int, List[int]]] = None,
        dummy_molecule: Optional[Molecule] = None,
        dummy_molecule_graph: Optional[MoleculeGraph] = None,
        dummy_molecule_indices_mapping: Optional[Dict[int, List[int]]] = None,
        dummy_branching_indices: Optional[Collection[int]] = None,
    ):
        """Initialize a secondary building block.

        In practice, you won't use this constructor directly.

        Args:
            molecule (Molecule): Pymatgen molecule object.
            molecule_graph (MoleculeGraph): Pymatgen molecule graph object.
            graph_branching_indices (Collection[int]): Branching indices
                in the original structure.
            binding_indices (Collection[int]): Binding indices in the original structure.
            molecule_original_indices_mapping (Optional[Dict[int, List[int]]], optional):
                Mapping from molecule indices to original indices. Defaults to None.
            dummy_molecule (Optional[Molecule], optional): Dummy molecule. Defaults to None.
            dummy_molecule_graph (Optional[MoleculeGraph], optional): Dummy molecule graph.
                Defaults to None.
            dummy_molecule_indices_mapping (Optional[Dict[int, List[int]]], optional):
                Dummy molecule indices mapping. Defaults to None.
            dummy_branching_indices (Optional[Collection[int]], optional):
                Dummy branching indices. Defaults to None.
        """
        self.molecule = molecule
        self._mapping = molecule_original_indices_mapping

        self._indices = sum(list(molecule_original_indices_mapping.values()), [])

        self._original_indices = self._indices
        self.molecule_graph = molecule_graph
        self._original_graph_branching_indices = graph_branching_indices
        self._original_binding_indices = binding_indices

        self._dummy_molecule = dummy_molecule
        self._dummy_molecule_graph = dummy_molecule_graph
        self._dummy_molecule_indices_mapping = dummy_molecule_indices_mapping
        self._dummy_branching_indices = dummy_branching_indices

        self.mapping_from_original_indices = defaultdict(list)
        if molecule_original_indices_mapping is None:
            for ori_index, index in zip(self._indices, range(len(molecule))):
                self.mapping_from_original_indices[ori_index].append(index)
        else:
            for k, v in molecule_original_indices_mapping.items():
                for i in v:
                    self.mapping_from_original_indices[i].append(k)

        if dummy_molecule:
            self.dummy_mapping_from_original_indices = defaultdict(list)
            if dummy_molecule_indices_mapping is None:
                for ori_index, index in zip(self._indices, range(len(dummy_molecule))):
                    self.dummy_mapping_from_original_indices[ori_index].append(index)
            else:
                for k, v in dummy_molecule_indices_mapping.items():
                    for i in v:
                        self.dummy_mapping_from_original_indices[i].append(k)

        self.mapping_to_original_indices = {}
        for key, value in self.mapping_from_original_indices.items():
            for v in value:
                self.mapping_to_original_indices[v] = key

    @property
    def center(self):
        return self.molecule.center_of_mass

    def search_pubchem(self, listkey_counts: int = 10, **kwargs) -> Tuple[List[str], bool]:
        """Search for a molecule in pubchem # noqa: DAR401

        Second element of return tuple is true if there was an identity match

        Args:
            listkey_counts (int): Number of list keys to return
                (relevant for substructure search).
                Defaults to 10.
            kwargs: Additional arguments to pass to PubChem.search

        Returns:
            Tuple[List[str], bool]: List of pubchem ids and whether there was an identity match
        """
        try:
            matches = pcp.get_compounds(
                self.smiles, namespace="smiles", searchtype="fastidentity", **kwargs
            )
            if matches:
                return (
                    matches,
                    True,
                )
            else:
                raise ValueError("No matches found")
        except Exception:
            logger.warning(
                f"Could not find {self.smiles} in pubchem, \
                    now performing substructure search"
            )
            # we use `fastsubstructure` as it fixes
            # https://github.com/kjappelbaum/moffragmentor/issues/63
            res = pcp.get_compounds(
                self.smiles,
                namespace="smiles",
                searchtype="fastsubstructure",
                listkey_counts=listkey_counts,
                **kwargs,
            )
            smiles = [r.canonical_smiles for r in res]

            return mcs_rank(self.smiles, smiles, additional_attributes=res), False

    def get_neighbor_indices(self, site: int) -> List[int]:
        """Get list of indices of neighboring sites"""
        return [site.index for site in self.molecule_graph.get_connected_sites(site)]

    # make this function so we can have different flavors
    def get_indices(self):
        return self._indices

    @property
    def is_edge(self):
        return len(self.branching_coords) == 2

    def __len__(self):
        """Return the number of atoms in the molecule."""
        return len(self.molecule)

    def __str__(self):
        """Return the SMILES string for the molecule."""
        return self.smiles

    def dump(self, path):
        pickle_dump(self, path)

    def _get_nx_graph(self):
        return nx.Graph(self.molecule_graph.graph.to_undirected())

    @cached_property
    def metal_indices(self) -> List[int]:
        return [i for i, species in enumerate(self.molecule.species) if species.is_metal]

    @cached_property
    def nx_graph(self):
        return self._get_nx_graph()

    @property
    def composition(self):
        return self.molecule.composition.alphabetical_formula

    @property
    def cart_coords(self):
        return self.molecule.cart_coords

    @cached_property
    def mol_with_coords(self):
        mol = self.molecule.copy()
        mol.coords = self.cart_coords
        sites = mol.sites
        return Molecule.from_sites(sites)

    @property
    def coordination(self):
        return len(self.original_graph_branching_indices)

    @property
    def original_graph_branching_indices(self):
        return self._original_graph_branching_indices

    @property
    def graph_branching_indices(self):
        indices = []
        if self._dummy_branching_indices is None:
            for i in self.original_graph_branching_indices:
                for index in self.mapping_from_original_indices[i]:
                    indices.append(index)
        else:
            for i in self._dummy_branching_indices:
                for index in self.dummy_mapping_from_original_indices[i]:
                    indices.append(index)
        return indices

    @cached_property
    def weisfeiler_lehman_graph_hash(self):
        return nx.weisfeiler_lehman_graph_hash(self.molecule_graph.graph, node_attr="specie")

    @property
    def molar_mass(self):
        return self.molecule.composition.weight

    @cached_property
    def hash(self) -> str:
        """Return hash.

        The hash is a combination of Weisfeiler-Lehman graph hash and center.

        Returns:
            str: Hash.
        """
        wl_hash = self.weisfeiler_lehman_graph_hash
        center = self.molecule.cart_coords.mean(axis=0)
        return f"{wl_hash}-{center[0]:.2f}-{center[1]:.2f}-{center[2]:.2f}"

    def __eq__(self, other: "SBU") -> bool:
        """Check if two molecules are equal.

        Based on the Weisfeiler-Lehman graph hash and center of mass.

        Args:
            other (SBU): SBU to compare to.

        Returns:
            bool: True if equal, False otherwise.
        """
        if hash(self) != hash(other):
            return False
        return True

    @cached_property
    def branching_coords(self):
        return (
            self.cart_coords[self.graph_branching_indices]
            if self._dummy_branching_indices is None
            else self._dummy_molecule.cart_coords[self.graph_branching_indices]
        )

    @property
    def original_binding_indices(self):
        return self._original_binding_indices

    @cached_property
    def binding_indices(self):
        indices = []
        for i in self.original_binding_indices:
            for index in self.mapping_from_original_indices[i]:
                indices.append(index)
        return indices

    @cached_property
    def rdkit_mol(self):
        return obmol_to_rdkit_mol(self.openbabel_mol)

    @cached_property
    def openbabel_mol(self):
        return self.get_openbabel_mol()

    def get_openbabel_mol(self):
        from openbabel import pybel as pb

        a = BabelMolAdaptor(self.molecule)
        pm = pb.Molecule(a.openbabel_mol)
        return pm

    def show_molecule(self):
        import nglview

        return nglview.show_pymatgen(self.molecule)

    def show_connecting_structure(self):
        import nglview

        return nglview.show_pymatgen(self._get_connected_sites_structure())

    def show_binding_structure(self):
        import nglview

        return nglview.show_pymatgen(self._get_binding_sites_structure())

    def to(self, fmt: str, filename: str):
        return self.molecule.to(fmt, filename)

    @cached_property
    def smiles(self) -> str:
        """Return canonical SMILES.

        Use openbabel to compute the SMILES, but then get the
        canonical version with RDKit as we observed sometimes the same
        molecule ends up as different canonical SMILES for openbabel.
        If RDKit cannot make a canonical SMILES (can happen with organometallics)
        we simply use the openbabel version.

        Returns:
            str: Canonical SMILES
        """
        mol = self.openbabel_mol
        smiles = mol.write("can").strip()
        try:
            canonical = Chem.CanonSmiles(smiles)
            return canonical
        except Exception:
            return smiles

    def _get_boxed_structure(self):
        max_size = _get_max_sep(self.molecule.cart_coords)
        structure = self.molecule.get_boxed_structure(
            max_size + 0.1 * max_size,
            max_size + 0.1 * max_size,
            max_size + 0.1 * max_size,
            reorder=False,
        )
        return structure

    def _get_binding_sites_structure(self):
        sites = []
        s = self._get_boxed_structure()
        for i in self.binding_indices:
            sites.append(s[i])
        return Structure.from_sites(sites)

    @cached_property
    def descriptors(self):
        return self._get_descriptors()


def _get_max_sep(coordinates):
    if len(coordinates) > 1:
        distances = pdist(coordinates)
        return np.max(distances)
    return 5
